"""
What's new in CF-1.8
--------------------
2.7. Groups
  2.7.1. Scope
  2.7.2. Application of attributes

6.1.2. Taxon Names and Identifiers

7.5. Geometries
"""

import itertools
import re
import warnings
from collections import defaultdict

import numpy as np
import requests
from lxml import etree
from netCDF4 import Dataset
from shapely.geometry import Polygon

from compliance_checker.base import BaseCheck, TestCtx
from compliance_checker.cf.cf_1_7 import CF1_7Check
from compliance_checker.cf.util import reference_attr_variables, string_from_var_type


class CF1_8Check(CF1_7Check):
    """Implementation for CF v1.8. Inherits from CF1_7Check."""

    # things that are specific to 1.8
    _cc_spec_version = "1.8"
    _cc_url = "http://cfconventions.org/Data/cf-conventions/cf-conventions-1.8/cf-conventions.html"

    ROOT_GROUP_ONLY_ATTRS = ["Conventions", "external_variables"]
    NON_ROOT_GROUP_OPT = ["title", "history"]

    def __init__(self, options=None):
        super().__init__(options)
        self.section_titles.update(
            {
                "2.7": "§2.7 Groups",
                "6.1.2": "§6.1.2 Taxon Names and Identifiers",
                "7.5": "§7.5 Geometries",
            },
        )

    def check_groups(self, ds: Dataset):
        """
        2.7.2. Application of attributes

        The following attributes are optional for non-root groups. They are allowed in order to
        provide additional provenance and description of the subsidiary data. They do not override
        attributes from parent groups.

        - title
        - history

        If these attributes are present, they may be applied additively to the parent attributes of
        the same name. If a file containing groups is modified, the user or application need only
        update these attributes in the root group, rather than traversing all groups and updating
        all attributes that are found with the same name. In the case of conflicts, the root group
        attribute takes precedence over per-group instances of these attributes.

        The following attributes MAY ONLY be used in the root group and SHALL NOT be duplicated or
        overridden in child groups:

        - Conventions
        - external_variables

        Furthermore, per-variable attributes MUST be attached to the variables to which they refer.
        They MAY NOT be attached to a group, even if all variables within that group use the same
        attribute and value.

        If attributes are present within groups without being attached to a variable, these
        attributes apply to the group where they are defined, and to that group’s descendants, but
        not to ancestor or sibling groups. If a group attribute is defined in a parent group, and
        one of the child group redefines the same attribute, the definition within the child group
        applies for the child and all of its descendants.
        """

        results = []

        ctx_hi = TestCtx(BaseCheck.HIGH, self.section_titles["2.7"])
        ctx_lo = TestCtx(BaseCheck.LOW, self.section_titles["2.7"])

        # IMPLEMENTATION CONFORMANCE 2.7 REQUIRED 1/4
        # Make sure `Conventions` & `external_variables` attributes are only present in the
        # root group.
        for gname in ds.groups:
            ginstance = ds.groups[gname]

            for attr in ginstance.ncattrs():
                if attr in CF1_8Check.ROOT_GROUP_ONLY_ATTRS:
                    ctx_hi.messages.append(
                        f'§2.7.2 Attribute "{attr}" MAY ONLY be used in the root group '
                        "and SHALL NOT be duplicated or overridden in child groups.",
                    )

                    results.append(ctx_hi.to_result())

                elif attr in CF1_8Check.NON_ROOT_GROUP_OPT:
                    ctx_lo.messages.append(
                        f'§2.7.2 Note: attribute "{attr}" found on non-root group "{gname}". '
                        "This is optional for non-root groups. It is allowed in order to provide additional "
                        "provenance and description of the subsidiary data. It does not override "
                        "attributes from parent groups.",
                    )
                    results.append(ctx_lo.to_result())

        return results

    def check_invalid_same_named_dimension_across_groups(self, ds):
        """
        Section 2 NetCDF Files and Components
        Section 2.7. Groups
        Section 2.7.1. Scope

        "If any dimension of an out-of-group variable has the same name as a dimension of
        the referring variable, the two must be the same dimension (i.e. they must have
        the same netCDF dimension ID)."
        """

        results = []

        invalid_same_named_dimension_across_groups = TestCtx(
            BaseCheck.HIGH,
            self.section_titles["2.7.1"],
        )

        # Using the group B dimension to match a variable from group A.
        # This is logically wrong: NetCDF allows it, but the dimensions are different objects.
        # This may not throw during creation, so we simulate an access/mismatch
        # grab the first group's time‐dim
        grp_list = list(ds.groups.keys())
        # Only run the assertion if there are at least two groups
        if len(grp_list) >= 2:
            # compare everyone to the first
            first_dim = ds.groups[grp_list[0]].dimensions["time"]

            # check that every other group's time‐dim is the very same object
            invalid_same_named_dimension_across_groups.assert_true(
                all(ds.groups[g].dimensions["time"] is first_dim for g in grp_list[1:]),
                "Dimensions with the same name must be the same object (ID).",
            )
            results.append(invalid_same_named_dimension_across_groups.to_result())

        return results

    def check_coordinates_with_paths(self, ds):
        """
        Section 2 NetCDF Files and Components
        Section 2.7. Groups
        Section 2.7.1. Scope

        Search by absolute path
        A variable or dimension specified with an absolute path (i.e., with a leading slash "/")
        is at the indicated location relative to the root group, as in a UNIX-style file convention.
        For example, a coordinates attribute of /g1/lat refers to the lat variable in group /g1.

        Search by relative path
        As in a UNIX-style file convention, a variable or dimension specified with a relative path
        (i.e., containing a slash but not with a leading slash, e.g. child/lat) is at the location
        obtained by affixing the relative path to the absolute path of the referring attribute.
        For example, a coordinates attribute of g1/lat refers to the lat variable in subgroup g1
        of the current (referring) group. Upward path traversals from the current group are indicated
        with the UNIX convention. For example, ../g1/lat refers to the lat variable in the sibling group
        g1 of the current (referring) group.

        Search by proximity
        A variable or dimension specified with no path (for example, lat) refers to the variable or
        dimension of that name, if there is one, in the referring group. If not, the ancestors of the
        referring group are searched for it, starting from the direct ancestor and proceeding toward
        the root group, until it is found.
        A special case exists for coordinate variables. Because coordinate variables must share dimensions
        with the variables that reference them, the ancestor search is executed only until the local apex
        group is reached. For coordinate variables that are not found in the referring group or its ancestors,
        a further strategy is provided, called lateral search. The lateral search proceeds downwards from the
        local apex group width-wise through each level of groups until the sought coordinate is found.
        The lateral search algorithm may only be used for NUG coordinate variables; it shall not be used for
        auxiliary coordinate variables.
        """
        results = []

        hi_detect_coordinates_paths = TestCtx(
            BaseCheck.HIGH,
            self.section_titles["2.7.1"],
        )
        lo_detect_coordinates_paths = TestCtx(
            BaseCheck.LOW,
            self.section_titles["2.7.1"],
        )

        # Walk the group tree from root
        group_stack = [("/", ds)]

        while group_stack:
            current_path, group = group_stack.pop()

            for var_name, var in group.variables.items():
                coords = (
                    var.getncattr("coordinates")
                    if "coordinates" in var.ncattrs()
                    else ""
                )
                for coord_ref in coords.split():

                    # ---------------------------------------
                    # 1. Absolute Path Check: /group1/lat
                    # ---------------------------------------
                    if coord_ref.startswith("/"):
                        try:
                            coord_var = ds
                            for part in coord_ref.strip("/").split("/")[:-1]:
                                coord_var = coord_var.groups[part]
                            coord_var.variables[coord_ref.split("/")[-1]]
                            found = True
                            hi_detect_coordinates_paths.messages.append(
                                f"{current_path}{var_name}: '{coord_ref}' found found (absolute path)",
                            )
                        except Exception:
                            hi_detect_coordinates_paths.messages.append(
                                f"{current_path}{var_name}: '{coord_ref}' not found at absolute path",
                            )

                    # ---------------------------------------
                    # 2. Relative Path Check: ../group1/lat
                    # ---------------------------------------
                    elif coord_ref.startswith("../") or coord_ref.startswith("./"):
                        try:
                            coord_var = group
                            for part in coord_ref.split("/")[:-1]:
                                if part == "..":
                                    coord_var = coord_var.parent
                                elif part == ".":
                                    continue
                                else:
                                    coord_var = coord_var.groups[part]
                            coord_var.variables[coord_ref.split("/")[-1]]
                            hi_detect_coordinates_paths.messages.append(
                                f"{current_path}{var_name}: '{coord_ref}' found (relative path)",
                            )
                        except Exception:
                            hi_detect_coordinates_paths.messages.append(
                                f"{current_path}{var_name}: '{coord_ref}' not found at relative path",
                            )

                    # ---------------------------------------
                    # 3. No Path (Unqualified name): 'lat'
                    #    → Proximity search: current group + ancestors
                    # ---------------------------------------
                    else:
                        # 3a: Check current group
                        if coord_ref in group.variables:
                            hi_detect_coordinates_paths.messages.append(
                                f"{current_path}{var_name}: '{coord_ref}' found in same group",
                            )
                            continue

                        # 3b. Search ancestor groups
                        parent = group
                        found = False
                        while hasattr(parent, "parent") and parent.parent is not None:
                            parent = parent.parent
                            if coord_ref in parent.variables:
                                hi_detect_coordinates_paths.messages.append(
                                    f"{current_path}{var_name}: '{coord_ref}' found in ancestor",
                                )
                                found = True
                                break

                        # 3c. Not found by proximity → likely lateral search
                        if not found:
                            hi_detect_coordinates_paths.messages.append(
                                f"{current_path}{var_name}: '{coord_ref}' not found in group or ancestors. It is likely a lateral search, which is discouraged",
                            )

                    # ---------------------------------------
                    # Extra Check: Is it a NUG coordinate variable?
                    # (Name == dimension name → lat(lat))
                    # CF discourages lateral search in this case
                    # ---------------------------------------
                    nug_coord_violation = False

                    def check_nug_variable(g, coord=coord_ref):
                        nonlocal nug_coord_violation
                        # Step 1: Check this group's variables
                        for vname, v in g.variables.items():
                            if (
                                vname == coord
                                and len(v.dimensions) == 1
                                and vname == v.dimensions[0]
                            ):
                                nug_coord_violation = True

                        # Step 2: Recursively check child groups
                        for sub in g.groups.values():
                            check_nug_variable(sub, coord)

                    check_nug_variable(ds, coord_ref)

                    if nug_coord_violation:
                        lo_detect_coordinates_paths.messages.append(
                            f" WARNING: '{coord_ref}' is a NUG coordinate variable used via lateral search. CF Recommendation: Use an absolute or relative path instead.",
                        )

            # Add child groups to the stack to continue walking the tree
            for name, subgroup in group.groups.items():
                group_stack.append((current_path + name + "/", subgroup))

        results.append(hi_detect_coordinates_paths.to_result())
        results.append(lo_detect_coordinates_paths.to_result())

        return results

    def check_geometry(self, ds: Dataset):
        """Runs any necessary checks for geometry well-formedness
        :param netCDF4.Dataset ds: An open netCDF dataset
        :returns list: List of error messages
        """
        results = []
        geom_valid = TestCtx(BaseCheck.MEDIUM, self.section_titles["7.5"])

        vars_with_geometry = ds.get_variables_by_attributes(
            geometry=lambda g: g is not None,
        )

        unique_geometry_var_names = defaultdict(list)
        for var in vars_with_geometry:
            unique_geometry_var_names[var.geometry].append(var)

        for geometry_var_name in unique_geometry_var_names:
            geom_valid.out_of += 1

            # IMPLEMENTATION CONFORMANCE 7.5 REQUIRED 2/20
            # The geometry container variable name must exist in the file.
            if geometry_var_name not in ds.variables:
                geom_valid.messages.append(
                    f"Cannot find geometry variable named {geometry_var_name}",
                )
                results.append(geom_valid.to_result())
                return results
                continue

            geometry_var = ds.variables[geometry_var_name]

            # IMPLEMENTATION CONFORMANCE 7.5 REQUIRED 10/20
            # The grid_mapping and coordinates attributes can be carried by the
            # geometry container variable provided they are also carried by the
            # data variables associated with the container.
            if hasattr(geometry_var, "coordinates"):
                coord_err_template = (
                    "Geometry variable {geometry_var_name} has attribute "
                    "coordinates which is either not present or is not a "
                    "subset of the coordinates attribute of "
                    "the referring parent variable {parent_var_name}"
                )
                # allow for geometry coordinates attribute to be a subset of
                # the parent variable's coordinates
                geom_var_coords = set(re.split(r"\s+", geometry_var.coordinates))
                for parent_var in vars_with_geometry:
                    geom_valid.out_of += 1
                    if not hasattr(parent_var, "coordinates"):
                        parent_var_coords = set()
                    else:
                        parent_var_coords = set(
                            re.split(r"\s+", parent_var.coordinates),
                        )
                    if not geom_var_coords.issubset(parent_var_coords):
                        err_msg = coord_err_template.format(
                            geometry_var_name=geometry_var_name,
                            parent_var_name=parent_var.name,
                        )
                        geom_valid.messages.append(err_msg)
                    else:
                        geom_valid.score += 1

            # IMPLEMENTATION CONFORMANCE 7.5 REQUIRED 10/20:
            # continued: grid_mapping match
            # TODO: should grid_mapping checks not be strict equality?
            if hasattr(geometry_var, "grid_mapping"):
                for parent_var in vars_with_geometry:
                    geom_valid.out_of += 1
                    grid_mapping_err_template = (
                        "Geometry variable {geometry_var_name} has "
                        "attribute grid_mapping which is "
                        "either not present or does not have the "
                        "same value as the referring parent variable "
                        "{parent_var_name}"
                    )
                    if geometry_var.grid_mapping != getattr(
                        parent_var,
                        "grid_mapping",
                        None,
                    ):
                        err_msg = grid_mapping_err_template.format(
                            geometry_var_name=geometry_var_name,
                            parent_var_name=parent_var.name,
                        )
                        geom_valid.messages.append(err_msg)
                    else:
                        geom_valid.score += 1

            # IMPLEMENTATION CONFORMANCE 7.5 REQUIRED 3/20
            # geometry container must have geometry_type
            geometry_type = getattr(geometry_var, "geometry_type", "").lower()

            # IMPLEMENTATION CONFORMANCE 7.5 REQUIRED 4/20:
            # only legal values for geometry_type
            if geometry_type not in ["point", "line", "polygon"]:
                geom_type_err_template = (
                    "geometry_type on {geometry_var_name}"
                    " must be 'point', 'line', or 'polygon'"
                )
                err_msg = geom_type_err_template.format(
                    geometry_var_name=geometry_var_name,
                )
                geom_valid.messages.append(err_msg)

                results.append(geom_valid.to_result())
                return results
                continue

            try:
                node_coord_var_names = geometry_var.node_coordinates
            except AttributeError:
                # IMPLEMENTATION CONFORMANCE 7.5 REQUIRED 3/20 continued:
                # geometry container must have node_coordinates
                geom_valid.messages.append(
                    "Could not find required attribute "
                    '"node_coordinates" in geometry '
                    f'variable "{geometry_var_name}"',
                )
                results.append(geom_valid.to_result())
                return results
                continue

            # IMPLEMENTATION CONFORMANCE 7.5 REQUIRED 7/20:
            # node_coordinates must be string
            if not isinstance(node_coord_var_names, str):
                geom_valid.messages.append(
                    "Attribute node_coordinates in geometry "
                    f'variable "{geometry_var_name}" must be '
                    "a string",
                )
                results.append(geom_valid.to_result())
                return results
                continue

            split_coord_names = node_coord_var_names.strip().split(" ")
            node_coord_vars, not_found_node_vars = [], []

            for coord_var_name in split_coord_names:
                try:
                    node_coord_vars.append(ds.variables[coord_var_name])
                except KeyError:
                    not_found_node_vars.append(coord_var_name)

            # IMPLEMENTATION CONFORMANCE 7.5 REQUIRED 7/20 continued:
            # all variable names must exist
            if not_found_node_vars:
                geom_valid.messages.append(
                    "The following referenced node coordinate "
                    "variables for geometry variable "
                    f"{geometry_var_name} were not found: "
                    f"{not_found_node_vars}",
                )
                results.append(geom_valid.to_result())
                return results
                continue

            # IMPLEMENTATION CONFORMANCE 7.5 REQUIRED 13/20:
            # All node coordinate variables must share the same single dimension
            node_coord_dims = set()
            for coord_var in node_coord_vars:
                if len(coord_var.dimensions) != 1:
                    geom_valid.messages.append(
                        f"Node coordinate var '{coord_var.name}' must have exactly one dimension",
                    )
                else:
                    node_coord_dims.add(coord_var.dimensions[0])

            if len(node_coord_dims) == 1:
                geom_valid.score += 1
                shared_geom_dim = list(node_coord_dims)[0]
            else:
                geom_valid.messages.append(
                    f"Node coordinate var '{coord_var.name}' must share the same single dimension",
                )
                shared_geom_dim = None

            # IMPLEMENTATION CONFORMANCE 7.5 REQUIRED 1/20:
            # At least one dim of parent variable must match geom dimension
            if shared_geom_dim:
                for parent_var in vars_with_geometry:
                    geom_valid.out_of += 1
                    if shared_geom_dim in parent_var.dimensions:
                        geom_valid.score += 1
                    else:
                        geom_valid.messages.append(
                            f"Parent variable '{parent_var.name}' does not include geometry "
                            f"dimension '{shared_geom_dim}' used in geometry variable '{geometry_var_name}'",
                        )

            # IMPLEMENTATION CONFORMANCE 7.5 REQUIRED 8–9/20:
            # All node coord vars must have axis, and they must be unique
            axes = []
            missing_axes = []
            for coord_var in node_coord_vars:
                geom_valid.out_of += 1
                if not hasattr(coord_var, "axis"):
                    missing_axes.append(coord_var.name)
                elif coord_var.axis not in ["X", "Y", "Z"]:
                    geom_valid.messages.append(
                        f"{coord_var.name} has an axis attribute whose allowable values are not X, Y, or Z",
                    )
                else:
                    geom_valid.score += 1
                    axes.append(coord_var.axis)

            if missing_axes:
                geom_valid.out_of += 1
                geom_valid.messages.append(
                    f"Missing axis attribute on node coord vars: {missing_axes}",
                )
            else:
                geom_valid.score += 1

            if len(set(axes)) != len(axes):
                geom_valid.out_of += 1
                geom_valid.messages.append(
                    f"Duplicate axis values among node coord vars: {axes}",
                )
            elif not missing_axes:
                geom_valid.score += 1

            # MPLEMENTATION CONFORMANCE 7.5 REQUIRED 11–12/20:
            # Validate nodes attribute on coordinate vars
            for coord_var in node_coord_vars:
                nodes_attr = getattr(coord_var, "nodes", None)
                if nodes_attr:
                    geom_valid.out_of += 1
                    if nodes_attr not in ds.variables:
                        geom_valid.messages.append(
                            f"'nodes' attr on {coord_var.name} references unknown var {nodes_attr}",
                        )
                    elif nodes_attr not in split_coord_names:
                        geom_valid.messages.append(
                            f"'nodes' attr on {coord_var.name} must point to one of the node coordinate vars",
                        )
                    else:
                        nodes_var = ds.variables[nodes_attr]
                        if (
                            hasattr(coord_var, "grid_mapping")
                            and hasattr(nodes_var, "grid_mapping")
                            and coord_var.grid_mapping != nodes_var.grid_mapping
                        ):
                            geom_valid.messages.append(
                                f"grid_mapping mismatch between {coord_var.name} and {nodes_attr}",
                            )
                        else:
                            geom_valid.score += 1

            # IMPLEMENTATION CONFORMANCE 7.5 REQUIRED 5–6/20
            # For a line geometry_type, each geometry must have a minimum of two node coordinates.
            # For a polygon geometry_type, each geometry must have a minimum of three node coordinates.

            if geometry_type == "line" and len(node_coord_vars[0]) < 2:
                geom_valid.out_of += 1
                geom_valid.messages.append(
                    f"Line geometry '{geometry_var_name}' must have ≥2 nodes",
                )
            elif geometry_type == "polygon" and len(node_coord_vars[0]) < 3:
                geom_valid.messages.append(
                    f"Polygon geometry '{geometry_var_name}' must have ≥3 nodes",
                )
            else:
                geom_valid.score += 1

            # IMPLEMENTATION CONFORMANCE 7.5 REQUIRED 14/20:
            # Nodes for polygon exterior rings must be put in anticlockwise order (viewed from above)
            # and polygon interior rings in clockwise order.
            # This is geometry-specific and may require geometric analysis,
            # so only warning is issued via geometry.check_geometry()

            # Load attributes
            node_count, node_count_errors = reference_attr_variables(
                ds,
                getattr(geometry_var, "node_count", None),
            )
            part_node_count, part_node_count_errors = reference_attr_variables(
                ds,
                getattr(geometry_var, "part_node_count", None),
            )
            interior_ring, interior_ring_errors = reference_attr_variables(
                ds,
                getattr(geometry_var, "interior_ring", None),
            )

            # Ensure variables are defined even if None is returned
            node_count = node_count if node_count is not None else None
            part_node_count = part_node_count if part_node_count is not None else None
            interior_ring = interior_ring if interior_ring is not None else None

            if geometry_type == "point":
                geometry = PointGeometry(node_coord_vars, node_count)
            elif geometry_type == "line":
                geometry = LineGeometry(node_coord_vars, node_count, part_node_count)
            elif geometry_type == "polygon":
                geometry = PolygonGeometry(
                    node_coord_vars,
                    node_count,
                    part_node_count,
                    interior_ring,
                )

            # IMPLEMENTATION CONFORMANCE 7.5 REQUIRED 15/20:
            # The single dimension of the part node count variable should equal the total number of parts in all the geometries.
            # polygons with interior geometry only
            if interior_ring is not None and part_node_count is not None:
                geom_valid.out_of += 1
                if len(interior_ring[:]) != len(part_node_count[:]):
                    geom_valid.messages.append(
                        f"part_node_count and interior_ring must have same length for '{geometry_var_name}'",
                    )
                else:
                    geom_valid.score += 1

            # IMPLEMENTATION CONFORMANCE 7.5 REQUIRED 16/20:
            # When node_count is missing and >1 node, type must be point
            try:
                length = len(node_coord_vars[0])
            except TypeError:
                length = getattr(node_coord_vars[0], "size", 1)  # fallback

            if not node_count and length > 1:
                geom_valid.out_of += 1
                if geometry_type != "point":
                    geom_valid.messages.append(
                        f"'{geometry_var_name}' missing node_count; must be point geometry",
                    )
                else:
                    geom_valid.score += 1

            # IMPLEMENTATION CONFORMANCE 7.5 REQUIRED 17/20:
            # If both node_count and part_node_count are present, their sums must match
            if part_node_count is not None and node_count is not None:
                geom_valid.out_of += 1
                if np.sum(part_node_count[:]) != np.sum(node_count[:]):
                    geom_valid.messages.append(
                        f"Sum mismatch: node_count = {np.sum(node_count[:])}, part_node_count = {np.sum(part_node_count[:])}",
                    )
                else:
                    geom_valid.score += 1

            # IMPLEMENTATION CONFORMANCE 7.5 REQUIRED 18/20:
            # If interior_ring attribute is present, part_node_count must also be present
            if interior_ring:
                geom_valid.out_of += 1
                if not part_node_count:
                    geom_valid.messages.append(
                        f"For geometry variable '{geometry_var_name}' which defines the interior_ring attribute, the part_node_count attribute must also be present",
                    )

                # IMPLEMENTATION CONFORMANCE 7.5 REQUIRED 19/20:
                # The single dimension of the interior_ring variable must match that of part_node_count
                elif (
                    part_node_count.dimensions != interior_ring.dimensions
                    or not len(part_node_count) == 1
                ):
                    geom_valid.messages.append(
                        f"part_node_count variable {part_node_count.name} must have the same single dimension as interior ring variable {interior_ring.name}",
                    )
                else:
                    geom_valid.score += 1

            # IMPLEMENTATION CONFORMANCE 7.5 REQUIRED 20/20:
            # The interior_ring variable must contain only 0 or 1 values
            if interior_ring is not None:
                geom_valid.out_of += 1
                unique_vals = np.unique(interior_ring[:])
                # Handle masked arrays safely
                if np.ma.isMaskedArray(unique_vals):
                    unique_vals = unique_vals.compressed()  # removes masked values

                if not set(unique_vals).issubset({0, 1}):
                    geom_valid.messages.append(
                        f"interior_ring variable {interior_ring.name} must contain only 0 (exterior) and 1 (interior) values",
                    )
                else:
                    geom_valid.score += 1

            # FINAL GEOMETRY VALIDATION:
            # geometric rules like ring orientation, intersections, etc.
            messages = geometry.check_geometry()
            if not messages:
                geom_valid.score += 1
            else:
                geom_valid.messages.extend(messages)

            results.append(geom_valid.to_result())
        return results

    def check_taxa(self, ds: Dataset):
        """
        6.1.2. Taxon Names and Identifiers

        A taxon is a named level within a biological classification, such as a class, genus and
        species. QUANTITIES DEPENDENT ON TAXA HAVE GENERIC STANDARD NAMES CONTAINING THE PHRASE
        "organisms_in_taxon", AND THE TAXA ARE IDENTIFIED BY AUXILIARY COORDINATE VARIABLES.

        The taxon auxiliary coordinate variables are string-valued. The plain-language name of the
        taxon MUST be contained in a variable with standard_name of 'biological_taxon_name'. A Life
        Science Identifier (LSID) may be contained in a variable with standard_name of
        biological_taxon_lsid. This is a URN with the syntax
        "urn:lsid:<Authority>:<Namespace>:<ObjectID>[:<Version>]". This includes the reference
        classification in the <Authority> element and these are restricted by the LSID governance.
        It is strongly recommended in CF that the authority chosen is World Register of Marine
        Species (WoRMS) for oceanographic data and Integrated Taxonomic Information System (ITIS)
        for freshwater and terrestrial data. WoRMS LSIDs are built from the WoRMS AphiaID taxon
        identifier such as "urn:lsid:marinespecies.org:taxname:104464" for AphiaID 104464. This may
        be converted to a URL by adding prefixes such as http://www.lsid.info/. ITIS LSIDs are
        built from the ITIS Taxonomic Serial Number (TSN), such as
        "urn:lsid:itis.gov:itis_tsn:180543".

        The biological_taxon_name auxiliary coordinate variable included for human readability is
        MANDATORY. The biological_taxon_lsid auxliary coordinate variable included for software
        agent readability is optional, but strongly recommended. If both are present then each
        biological_taxon_name coordinate must exactly match the name resolved from the
        biological_taxon_lsid coordinate. If LSIDs are available for some taxa in a dataset then
        the biological_taxon_lsid auxiliary coordinate variable should be included and missing data
        given for those taxa that do not have an identifier.
        """
        ret_val = []

        def match_taxa_standard_names(standard_name_string):
            """
            Match variables which are standard_names related to taxa, but
            are not the taxon identifiers or LSIDs themselves.
            """
            return (
                standard_name_string is not None
                and "taxon" in standard_name_string
                and standard_name_string  # exclude the identifiers we just looked at
                not in {"biological_taxon_lsid", "biological_taxon_name"}
                and standard_name_string in self._std_names
            )

        taxa_quantifier_variables = ds.get_variables_by_attributes(
            standard_name=match_taxa_standard_names,
        )
        # If there are no matches, there either are no taxa variables
        # or the standard names are not appropriate, which will be picked up
        # by the standard_name check
        if not taxa_quantifier_variables:
            return

        for taxon_quantifier_variable in taxa_quantifier_variables:
            valid_taxa = TestCtx(BaseCheck.HIGH, self.section_titles["6.1.2"])
            if not isinstance(
                getattr(taxon_quantifier_variable, "coordinates", None),
                str,
            ):
                valid_taxa.add_failure(
                    f'{taxon_quantifier_variable.name} must have a string valued "coordinates" attribute',
                )
                continue

            coordinate_var_names = taxon_quantifier_variable.coordinates.split(" ")
            invalid_coord_vars = set(coordinate_var_names) - ds.variables.keys()
            if invalid_coord_vars:
                valid_taxa.add_failure(
                    'The following values for "coordinates" attributes were not found in the dataset\'s variables '
                    f"{invalid_coord_vars}",
                )

            if len(coordinate_var_names) > 2:
                valid_taxa.add_failure(
                    "coordinates attribute for taxon data must either reference one or two variable names",
                )
                continue

            coordinate_vars = [
                ds.variables[var_name] for var_name in coordinate_var_names
            ]

            coord_var_standard_names = {
                var: getattr(var, "standard_name", None) for var in coordinate_vars
            }
            # if we have no authority, we can't check validity of the name -- assume it's OK
            standard_name_set = set(coord_var_standard_names.values())
            if set(coord_var_standard_names.keys()) == {"biological_taxon_name"}:
                # TODO: Check for at least binomial nomenclature?
                continue
            # check against WoRMS or ITIS if applicable
            elif standard_name_set == {
                "biological_taxon_name",
                "biological_taxon_lsid",
            }:
                inverted_dict = {v: k for k, v in coord_var_standard_names.items()}
                taxon_lsid_var = inverted_dict["biological_taxon_lsid"]
                taxon_name_var = inverted_dict["biological_taxon_name"]
                lsid_messages = self.handle_lsid(taxon_lsid_var, taxon_name_var)
                valid_taxa.out_of += 1
                if lsid_messages:
                    valid_taxa.messages.extend(lsid_messages)
                else:
                    valid_taxa.score += 1
            else:
                valid_taxa.add_failure(
                    f"coordinates attribute for variable {taxon_quantifier_variable} must consist of "
                    'variables containing standard names of either just "biological_taxon_name", or "biological_taxon_name" and "biological_taxon_identifier"',
                )
            ret_val.append(valid_taxa.to_result())

        return ret_val

    def handle_lsid(self, taxon_lsid_variable, taxon_name_variable):
        """
        Checks if LSID is well formed and present in the LSID database,
        and then attempts to delegate to WoRMS or ITIS, the LSID is applicable.
        If the LSID does not check the above authorities, it is not
        currently checked for correctness.
        """
        messages = []
        match_str = (
            r"(?:http://(?:www\.)?lsid.info/)?urn:lsid:"
            r"(?P<authority>[^:]+):(?P<namespace>[^:]+):"
            r"(?P<object_id>\w+)(?::(?P<version>\w+))?"
        )
        for taxon_lsid, taxon_name in zip(
            taxon_lsid_variable[:],
            taxon_name_variable[:],
            strict=False,
        ):
            # TODO: handle case where LSID is not present.  This can happen
            #       if the species is not present in the database desired.
            taxon_name_str = string_from_var_type(taxon_name)
            lsid_str = string_from_var_type(taxon_lsid)
            # if nodata/empty string for LSID, skip validity check
            if lsid_str == "":
                continue
            taxon_match = re.fullmatch(match_str, lsid_str)
            if not taxon_match:
                messages.append(
                    "Taxon id must match one of the following forms:\n"
                    "- urn:lsid:<authority>:<namespace>:<object_id>\n"
                    "- urn:lsid:<authority>:<namespace>:<object_id>:<version>\n"
                    "- www.lsid.info/urn:lsid.info:<authority>:<namespace>/<object_id>\n"
                    "- www.lsid.info/urn:lsid.info:<authority>:<namespace>/<object_id>:<version>\n"
                    "- lsid.info/urn:lsid.info:<authority>:<namespace>/<object_id>\n"
                    "- lsid.info/urn:lsid.info:<authority>:<namespace>/<object_id>:<version>\n"
                    "- http://lsid.info/urn:lsid.info:<authority>:<namespace>/<object_id>\n"
                    "- http://lsid.info/urn:lsid.info:<authority>:<namespace>/<object_id>:<version>\n"
                    "- http://www.lsid.info/urn:lsid.info:<authority>:<namespace>/<object_id>\n"
                    "- http://www.lsid.info/urn:lsid.info:<authority>:<namespace>/<object_id>:<version>",
                )
                continue
            if lsid_str.startswith("urn"):
                lsid_url = f"http://www.lsid.info/{lsid_str}"
            else:
                lsid_url = lsid_str

            try:
                response = requests.get(lsid_url, timeout=10)
                response.raise_for_status()
            except requests.exceptions.RequestException as e:
                # 400 error code indicates something is malformed on client's
                # end
                if response.status_code == 400:
                    tree = etree.HTML(response.text)
                    problem_text = tree.find("./body/p").text
                    messages.append(
                        "http://lsid.info returned an error message "
                        f"for submitted LSID string '{lsid_str}': "
                        f"{problem_text}",
                    )
                else:
                    messages.append(
                        "Error occurred attempting to check LSID "
                        f"'{lsid_str}': {str(e)}",
                    )
                continue

            # WoRMS -- marine bio data
            if (
                taxon_match["authority"] == "marinespecies.org"
                and taxon_match["namespace"] == "taxname"
            ):
                try:
                    response = requests.get(
                        f"http://www.marinespecies.org/rest/AphiaRecordByAphiaID/{taxon_match['object_id']}",
                        timeout=15,
                    )
                    response.raise_for_status()
                except requests.exceptions.RequestException as e:  # noqa: F841
                    messages.append(
                        "Aphia ID {taxon_match['object_id'] returned "
                        "other error: {str(e)}",
                    )
                # record not found in database
                if response.status_code == 204:
                    messages.append(
                        "Aphia ID {taxon_match['object_id'] "
                        "not found in WoRMS database",
                    )
                # good case, parse JSON
                elif response.status_code == 200:
                    valid_name = response.json()["valid_name"]
                    if valid_name != taxon_name_str:
                        messages.append(
                            "Supplied taxon name and WoRMS valid name do not match. "
                            f"Supplied taxon name is '{taxon_name_str}', WoRMS valid name "
                            f"is '{valid_name}.'",
                        )
                # Misc non-error code.  Should not reach here.
                else:
                    messages.append(
                        f"Aphia ID {taxon_match['object_id']}"
                        "returned an unhandled HTTP status "
                        f"code {response.status_code}",
                    )
                    continue

            # ITIS -- freshwater bio data
            elif (
                taxon_match["authority"] == "itis.gov"
                and taxon_match["namespace"] == "itis_tsn"
            ):
                itis_url = f"https://www.itis.gov/ITISWebService/jsonservice/getFullRecordFromTSN?tsn={taxon_match['object_id']}"
                try:
                    itis_response = requests.get(itis_url, timeout=15)
                    itis_response.raise_for_status()
                except requests.exceptions.RequestException as e:
                    if itis_response.status_code == 404:
                        messages.append(
                            "itis.gov TSN " f"{taxon_match['object_id']} not found.",
                        )
                        continue
                    else:
                        messages.append(
                            "itis.gov identifier returned other " f"error: {str(e)}",
                        )
                        continue
                json_contents = itis_response.json()
                combined_name = json_contents["scientificName"]["combinedName"]

                if taxon_name_str != combined_name:
                    messages.append(
                        "Supplied taxon name and ITIS scientific name do not match. "
                        f"Supplied taxon name is '{taxon_name_str}', ITIS scientific name "
                        f"for TSN {taxon_match['object_id']} is '{combined_name}.'",
                    )

            else:
                warnings.warn(
                    "Compliance checker only supports checking valid "
                    "LSID URNs of the form "
                    "'urn:lsid:marinespecies.org:taxname:<AphiaID>' or "
                    "'urn:lsid:itis.gov:itis_tsn:<TSN>'.  Assuming "
                    "pass condition",
                    stacklevel=1,
                )

        return messages


class GeometryStorage:
    """Abstract base class for geometries"""

    def __init__(self, coord_vars, node_count):
        self.coord_vars = coord_vars
        self.node_count = node_count
        self.errors = []
        # geometry is later parsed after sanity checks are run
        self.geometry = None

    def check_geometry(self):
        invalid_vars = []
        for coord_var in self.coord_vars:
            if not np.issubdtype(coord_var, float):
                invalid_vars.append(coord_var.name)
        # can't continue if the geometry variables are not the correct size
        if invalid_vars:
            self.errors.append(
                "The following geometry variables "
                f"have non-numeric contents: {invalid_vars}",
            )

    def _split_mulitpart_geometry(self):
        arr_extents_filt = self.part_node_count[self.part_node_count > 0]
        splits = np.split(np.vstack(self.coord_vars).T, arr_extents_filt.cumsum()[:-1])
        return splits


class PointGeometry(GeometryStorage):
    """Class for validating Point/MultiPoint geometries"""

    def check_geometry(self):
        super().check_geometry()

        if all(len(cv.dimensions) != 0 for cv in self.coord_vars):
            same_dim_group = itertools.groupby(self.coord_vars, lambda x: x.dimensions)
            same_dim = next(same_dim_group, True) and not next(same_dim_group, False)
            if not same_dim:
                self.errors.append(
                    "For a point geometry, coordinate "
                    "variables must be the same length as "
                    "node_count defined, or must be "
                    "length 1 if node_count is not set",
                )
        return self.errors


class LineGeometry(GeometryStorage):
    """Class for validating Line/MultiLine geometries"""

    def __init__(self, coord_vars, node_count, part_node_count):
        super().__init__(coord_vars, node_count)
        self.part_node_count = part_node_count
        if self.node_count is None or not np.issubdtype(
            self.node_count.dtype,
            np.integer,
        ):
            raise TypeError("For line geometries, node_count must be an integer")

    def check_geometry(self):
        geom_errors = []
        same_dim_group = itertools.groupby(self.coord_vars, lambda x: x.dimensions)
        same_dim = next(same_dim_group, True) and not next(same_dim_group, False)
        if not same_dim:
            raise IndexError(
                "Coordinate variables must be the same length. "
                "If node_count is specified, this value must "
                "also sum to the length of the coordinate "
                "variables.",
            )
        # if a multipart
        if self.node_count is not None:
            same_length = (
                len(self.coord_vars[0]) == np.atleast_1d(self.node_count)[:].sum()
            )
            if not same_length:
                geom_errors.append(
                    "Coordinate variables must be the same "
                    "length. If node_count is specified, this "
                    "value must also sum to the length of the "
                    "coordinate variables.",
                )
        if self.part_node_count is not None:
            if not np.issubdtype(self.part_node_count.dtype, np.integer):
                geom_errors.append(
                    "when part_node_count is specified, it must "
                    "be an array of integers",
                )
            same_node_count = len(self.coord_vars[0]) == self.node_count[:].sum()
            if not same_node_count:
                geom_errors.append(
                    "The sum of part_node_count must be equal "
                    "to the value of node_count",
                )
        return geom_errors


class PolygonGeometry(LineGeometry):
    """Class for validating Line/MultiLine geometries"""

    # TODO/clarify: Should polygons be simple, i.e. non-self intersecting?
    # Presumably
    def __init__(self, coord_vars, node_count, part_node_count, interior_ring):
        super().__init__(coord_vars, node_count, part_node_count)
        self.part_node_count = part_node_count
        self.interior_ring = interior_ring

    def check_polygon_orientation(self, transposed_coords, interior=False):
        """
        Checks that the polygon orientation is counter-clockwise if an
        exterior ring, otherwise clockwise if an interior ring.  Orientation
        is indicated by the `interior` boolean variable with False for an
        exterior ring and True for an interior ring (hole), defaulting to False.
        This function operates piecewise on individual interior/exterior
        polygons as well as multipart polygons
        :param np.array transposed_coords: A 2-by-n array of x and y coordinates
        :param bool interior: A boolean defaulting to False which has False
        indicating a counter-clockwise or exterior polygon, and True
        indicating a clockwise or interior polygon.
        :rtype bool:
        :returns: True if the polygon follows the proper orientation,
                  False if it fails the orientation test.
        """

        try:
            polygon = Polygon(transposed_coords.tolist())
        except ValueError as err:
            raise ValueError from err(
                "Polygon contains too few points to perform orientation test",
            )

        ccw = polygon.exterior.is_ccw
        return not ccw if interior else ccw

    def check_geometry(self):
        messages = super().check_geometry()
        # If any errors occurred within the preliminary checks, they preclude
        # running checks against the geometry here.
        if messages:
            return messages
        if self.part_node_count is not None:
            extents = np.concatenate([np.array([0]), self.part_node_count[:].cumsum()])
            if self.interior_ring is not None:
                ring_orientation = self.interior_ring[:].astype(bool)
            else:
                ring_orientation = np.zeros(len(self.part_count), dtype=bool)
            node_indexer_len = len(self.part_node_count)
        else:
            extents = np.concatenate([np.array([0]), self.node_count[:].cumsum()])
            node_indexer_len = len(self.node_count)
            ring_orientation = np.zeros(node_indexer_len, dtype=bool)
        # TODO: is it necessary to check whether part_node_count "consumes"
        #       node_count in the polygon, i.e. first (3, 3, 3) will consume
        #       a node part of 9, follow by next 3 will consume a node part of
        #       3 after consuming
        for i in range(node_indexer_len):
            extent_slice = slice(extents[i], extents[i + 1])
            poly_sliced = np.vstack([cv[extent_slice] for cv in self.coord_vars]).T
            pass_orientation = self.check_polygon_orientation(
                poly_sliced,
                ring_orientation[i],
            )
            if not pass_orientation:
                orient_fix = (
                    ("exterior", "counterclockwise")
                    if not ring_orientation[i]
                    else ("interior", "clockwise")
                )
                message = (
                    f"An {orient_fix[0]} polygon referred to by "
                    f"coordinates ({poly_sliced}) must have coordinates "
                    f"in {orient_fix[1]} order"
                )
                messages.append(message)
        return messages



#!/usr/bin/python

"""
Appendix E: Cell Methods
To be imported into cf.py upon initialization of a CF Checker class.
"""

cell_methods16 = {
    "point",
    "sum",
    "mean",
    "maximum",
    "minimum",
    "mid_range",
    "standard_deviation",
    "variance",
    "mode",
    "median",
    "sum_of_squares",
}

cell_methods17 = cell_methods16.union(
    {  # returns new set with elements from both
        "maximum_absolute_value",
        "minimum_absolute_value",
        "mean_absolute_value",
        "mean_of_upper_decile",
        "range",
        "root_mean_square",
    },
)



import logging
import os
import sqlite3
from warnings import warn

import numpy as np
import pyproj

import compliance_checker.cf.util as cfutil
from compliance_checker.base import BaseCheck, Result, TestCtx
from compliance_checker.cf.appendix_d import dimless_vertical_coordinates_1_7
from compliance_checker.cf.appendix_e import cell_methods17
from compliance_checker.cf.appendix_f import (
    ellipsoid_names17,
    grid_mapping_attr_types17,
    grid_mapping_dict17,
    horizontal_datum_names17,
    prime_meridian_names17,
)
from compliance_checker.cf.cf_1_6 import CF1_6Check
from compliance_checker.cf.cf_base import appendix_a_base

logger = logging.getLogger(__name__)


class CF1_7Check(CF1_6Check):
    """Implementation for CF v1.7. Inherits from CF1_6Check as most of the
    checks are the same."""

    # things that are specific to 1.7
    _cc_spec_version = "1.7"
    _cc_url = "http://cfconventions.org/Data/cf-conventions/cf-conventions-1.7/cf-conventions.html"

    appendix_a = appendix_a_base.copy()
    appendix_a.update(
        {
            "actual_range": {
                "Type": "N",
                "attr_loc": {"D", "C"},
                "cf_section": "2.5.1",
            },
            "comment": {
                "Type": "S",
                "attr_loc": {"G", "D", "C"},
                "cf_section": "2.6.2",
            },
            "external_variables": {
                "Type": "S",
                "attr_loc": {"G"},
                "cf_section": "2.6.3",
            },
            "scale_factor": {"Type": "N", "attr_loc": {"D", "C"}, "cf_section": "8.1"},
        },
    )

    def __init__(self, options=None):
        super().__init__(options)

        self.cell_methods = cell_methods17
        self.grid_mapping_dict = grid_mapping_dict17
        self.grid_mapping_attr_types = grid_mapping_attr_types17

    def check_external_variables(self, ds):
        """
        The global external_variables attribute is a blank-separated list of the
        names of variables which are named by attributes in the file but which
        are not present in the file. These variables are to be found in other
        files (called "external files") but CF does not provide conventions for
        identifying the files concerned. The only attribute for which CF
        standardises the use of external variables is cell_measures.

        :param netCDF4.Dataset ds: An open netCDF dataset
        :rtype: compliance_checker.base.Result
        """
        external_vars_ctx = TestCtx(BaseCheck.MEDIUM, self.section_titles["2.6.3"])
        # IMPLEMENTATION CONFORMANCE 2.6.3 REQUIRED 2/2
        try:
            # Should external variables be stored in dataset state?
            external_var_names = set(ds.external_variables.strip().split())

            bad_external_var_names = external_var_names.intersection(ds.variables)

            if bad_external_var_names:
                external_vars_ctx.out_of += 1
                bad_msg = (
                    "Global attribute external_variables should not "
                    "have any variable names which are present in the dataset. "
                    "Currently, the following names appear in both external_variables "
                    f"and the dataset's variables: {bad_external_var_names}"
                )
                external_vars_ctx.messages.append(bad_msg)

        # string/global attributes are handled in Appendix A checks
        except (AttributeError, ValueError):
            pass

        return external_vars_ctx.to_result()

    def check_actual_range(self, ds):
        """
        Check the actual_range attribute of variables. As stated in
        section 2.5.1 of version 1.7, this convention defines a two-element
        vector attribute designed to describe the actual minimum and actual
        maximum values of variables containing numeric data. Conditions:
          - the fist value of the two-element vector must be equal to the
            minimum of the data, and the second element equal to the maximum
          - if the data is packed, the elements of actual_range should have
            the same data type as the *unpacked* data
          - if valid_range is specified, both elements of actual_range should
            be within valid_range

        If a variable does not have an actual_range attribute, let it pass;
        including this attribute is only suggested. However, if the user is
        specifying the actual_range, the Result will be considered
        high-priority."""

        ret_val = []

        for name, variable in ds.variables.items():
            msgs = []
            score = 0
            out_of = 0

            if not hasattr(variable, "actual_range"):
                continue  # having this attr is only suggested, no Result needed
            else:
                out_of += 1
                try:
                    if (
                        len(variable.actual_range) != 2
                    ):  # TODO is the attr also a numpy array? if so, .size
                        msgs.append(
                            f"actual_range of '{name}' must be 2 elements",
                        )
                        ret_val.append(
                            Result(  # putting result into list
                                BaseCheck.HIGH,
                                (score, out_of),
                                self.section_titles["2.5"],
                                msgs,
                            ),
                        )
                        continue  # no need to keep checking if already completely wrong
                    else:
                        score += 1
                except TypeError:  # in case it's just a single number
                    msgs.append(f"actual_range of '{name}' must be 2 elements")
                    ret_val.append(
                        Result(  # putting result into list
                            BaseCheck.HIGH,
                            (score, out_of),
                            self.section_titles["2.5"],
                            msgs,
                        ),
                    )
                    continue

                # check equality to existing min/max values
                # NOTE this is a data check
                # If every value is masked, a data check of actual_range isn't
                # appropriate, so skip.
                if not (hasattr(variable[:], "mask") and variable[:].mask.all()):
                    # if min/max values aren't close to actual_range bounds,
                    # fail.
                    out_of += 1
                    if not np.isclose(
                        variable.actual_range[0],
                        variable[:].min(),
                    ) or not np.isclose(variable.actual_range[1], variable[:].max()):
                        msgs.append(
                            f"actual_range elements of '{name}' inconsistent with its min/max values",
                        )
                    else:
                        score += 1

                # check that the actual range is within the valid range
                if hasattr(variable, "valid_range"):  # check within valid_range
                    out_of += 1
                    if (variable.actual_range[0] < variable.valid_range[0]) or (
                        variable.actual_range[1] > variable.valid_range[1]
                    ):
                        msgs.append(
                            f'"{name}"\'s actual_range must be within valid_range',
                        )
                    else:
                        score += 1

                # check the elements of the actual range have the appropriate
                # relationship to the valid_min and valid_max
                if hasattr(variable, "valid_min"):
                    out_of += 1
                    if variable.actual_range[0] < variable.valid_min:
                        msgs.append(
                            f'"{name}"\'s actual_range first element must be >= valid_min ({variable.valid_min})',
                        )
                    else:
                        score += 1
                if hasattr(variable, "valid_max"):
                    out_of += 1
                    if variable.actual_range[1] > variable.valid_max:
                        msgs.append(
                            f'"{name}"\'s actual_range second element must be <= valid_max ({variable.valid_max})',
                        )
                    else:
                        score += 1

            ret_val.append(
                Result(  # putting result into list
                    BaseCheck.HIGH,
                    (score, out_of),
                    self.section_titles["2.5"],
                    msgs,
                ),
            )
        return ret_val

    def check_cell_boundaries(self, ds):
        """
        Checks the dimensions of cell boundary variables to ensure they are CF compliant
        per section 7.1.

        This method extends the CF1_6Check method; please see the original method for the
        complete doc string.

        If any variable contains both a formula_terms attribute *and* a bounding variable,
        that bounds variable must also have a formula_terms attribute.

        :param netCDF4.Dataset ds: An open netCDF dataset
        :returns list: List of results
        """

        # Note that test does not check monotonicity
        ret_val = []
        reasoning = []
        for variable_name, boundary_variable_name in cfutil.get_cell_boundary_map(
            ds,
        ).items():
            variable = ds.variables[variable_name]
            valid = True
            reasoning = []

            # 7.1 Required 1/5:
            # The type of the bounds attribute is a string whose value is a single variable name.
            # The specified variable must exist in the file.
            if boundary_variable_name not in ds.variables:
                valid = False
                reasoning.append(
                    f"Boundary variable {boundary_variable_name} referenced by {variable.name} not "
                    + "found in dataset variables",
                )
            else:
                boundary_variable = ds.variables[boundary_variable_name]

            # 7.1 Required 2/5:
            # The number of dimensions in the bounds variable should always be
            # the number of dimensions in the referring variable + 1
            if boundary_variable.ndim < 2:
                valid = False
                reasoning.append(
                    f"Boundary variable {boundary_variable.name} specified by {variable.name}"
                    + " should have at least two dimensions to enclose the base "
                    + "case of a one dimensionsal variable",
                )
            if boundary_variable.ndim != variable.ndim + 1:
                valid = False
                reasoning.append(
                    f"The number of dimensions of the variable {variable.name} is {variable.ndim}, but the "
                    f"number of dimensions of the boundary variable {boundary_variable.name} is {boundary_variable.ndim}. The boundary variable "
                    f"should have {variable.ndim + 1} dimensions",
                )
            if variable.dimensions[:] != boundary_variable.dimensions[: variable.ndim]:
                valid = False
                reasoning.append(
                    f"Boundary variable coordinates (for {variable.name}) are in improper order: {boundary_variable.dimensions}. Bounds-specific dimensions should be last"
                    "",
                )

            # 7.1 Required 2/5: continue
            # Ensure p vertices form a valid simplex given previous a...n
            # previous auxiliary coordinates
            if (
                ds.dimensions[boundary_variable.dimensions[-1]].size
                < len(boundary_variable.dimensions[:-1]) + 1
            ):
                valid = False
                reasoning.append(
                    f"Dimension {boundary_variable.name} of boundary variable (for {variable.name}) must have at least {len(variable.dimensions) + 1} elements to form a simplex/closed cell with previous dimensions {boundary_variable.dimensions[:-1]}.",
                )

            # 7.1 Required 3/5:
            # A boundary variable must be a numeric data type
            if boundary_variable.dtype.kind not in "biufc":
                valid = False
                reasoning.append(
                    f"Boundary variable {boundary_variable.name} specified by {variable.name}"
                    + "must be a numeric data type ",
                )

            # 7.1 Required 4/5:
            # If a boundary variable has units, standard_name, axis, positive, calendar, leap_month,
            # leap_year or month_lengths attributes, they must agree with those of its associated variable.
            if boundary_variable.__dict__.keys():
                for item in boundary_variable.__dict__.keys():
                    if hasattr(variable, item):
                        if getattr(variable, item) != getattr(boundary_variable, item):
                            valid = False
                            reasoning.append(
                                f"'{boundary_variable_name}' has attr '{item}' with value '{getattr(boundary_variable, item)}' that does not agree "
                                f"with its associated variable ('{variable.name}')'s attr value '{getattr(variable, item)}'"
                                "",
                            )

            # 7.1 Required 5/5:
            # check if formula_terms is present in the var; if so,
            # the bounds variable must also have a formula_terms attr
            if hasattr(variable, "formula_terms"):
                if not hasattr(boundary_variable, "formula_terms"):
                    valid = False
                    reasoning.append(
                        f"'{variable_name}' has 'formula_terms' attr, bounds variable '{boundary_variable_name}' must also have 'formula_terms'",
                    )

            # 7.1 Recommendations 2/2
            # Boundary variables should not have the _FillValue, missing_value, units, standard_name, axis,
            # positive, calendar, leap_month, leap_year or month_lengths attributes.
            attributes_to_check = {
                "_FillValue",
                "missing_value",
                "units",
                "standard_name",
                "axis",
                "positive",
                "calendar",
                "leap_month",
                "leap_year",
                "month_lengths",
            }
            if boundary_variable.__dict__.keys():
                lst1 = boundary_variable.__dict__.keys()
                lst2 = attributes_to_check
                unwanted_attributes = [value for value in lst1 if value in lst2]
                if unwanted_attributes:
                    valid = False
                    reasoning.append(
                        f"The Boundary variables '{boundary_variable_name}' should not have the attributes: '{unwanted_attributes}'",
                    )

            result = Result(
                BaseCheck.MEDIUM,
                valid,
                self.section_titles["7.1"],
                reasoning,
            )
            ret_val.append(result)
        return ret_val

    def check_cell_boundaries_interval(self, ds):
        """
        7.1 Cell Boundaries
        Recommendations: (1/2)
        The points specified by a coordinate or auxiliary coordinate variable
        should lie within, or on the boundary, of the cells specified by the
        associated boundary variable.
        """
        ret_val = []
        reasoning = []
        for variable_name, boundary_variable_name in cfutil.get_cell_boundary_map(
            ds,
        ).items():
            valid = True

            variable = ds.variables[variable_name]
            boundary_variable = ds.variables[boundary_variable_name]

            for ii in range(len(variable[:])):
                if abs(boundary_variable[ii][1]) >= abs(boundary_variable[ii][0]):
                    if not (
                        (abs(variable[ii]) >= abs(boundary_variable[ii][0]))
                        and (abs(variable[ii]) <= abs(boundary_variable[ii][1]))
                    ):
                        valid = False
                        reasoning.append(
                            f"The points specified by the coordinate variable {variable_name} ({variable[ii]})"
                            " lie outside the boundary of the cell specified by the "
                            f"associated boundary variable {boundary_variable_name} ({boundary_variable[ii]})",
                        )

                result = Result(
                    BaseCheck.MEDIUM,
                    valid,
                    self.section_titles["7.1"],
                    reasoning,
                )
                ret_val.append(result)
            return ret_val

    def check_cell_measures(self, ds):
        """
        A method to over-ride the CF1_6Check method. In CF 1.7, it is specified
        that variable referenced by cell_measures must be in the dataset OR
        referenced by the global attribute "external_variables", which represent
        all the variables used in the dataset but not found in the dataset.

        7.2 To indicate extra information about the spatial properties of a
        variable's grid cells, a cell_measures attribute may be defined for a
        variable. This is a string attribute comprising a list of
        blank-separated pairs of words of the form "measure: name". "area" and
        "volume" are the only defined measures.

        The "name" is the name of the variable containing the measure values,
        which we refer to as a "measure variable". The dimensions of the
        measure variable should be the same as or a subset of the dimensions of
        the variable to which they are related, but their order is not
        restricted.

        The variable must have a units attribute and may have other attributes
        such as a standard_name.

        :param netCDF4.Dataset ds: An open netCDF dataset
        :rtype: list
        :return: List of results
        """
        ret_val = []
        variables = ds.get_variables_by_attributes(
            cell_measures=lambda c: c is not None,
        )
        try:
            external_variables_str = ds.getncattr("external_variables")
            if external_variables_str is not None:
                external_variables_names = set(external_variables_str.split(" "))
            else:
                external_variables_names = set()
        except (ValueError, AttributeError):
            external_variables_names = set()
        for var in variables:
            ret_val.append(
                self._cell_measures_core(
                    ds,
                    var,
                    external_variables_names,
                    "dataset or external variable",
                ),
            )

        return ret_val

    def _check_grid_mapping_attr_condition(self, attr, attr_name):
        """
        Evaluate a condition (or series of conditions) for a particular
        attribute. Implementation for CF-1.7.

        :param attr: attribute to teset condition for
        :param str attr_name: name of the attribute
        :rtype tuple
        :return two-tuple of (bool, str)
        """

        if attr_name == "geographic_crs_name":
            return self._evaluate_geographic_crs_name(attr)

        elif attr_name == "geoid_name":
            return self._evaluate_geoid_name(attr)

        elif attr_name == "geopotential_datum_name":
            return self._evaluate_geopotential_datum_name(attr)

        elif attr_name == "horizontal_datum_name":
            return self._evaluate_horizontal_datum_name(attr)

        elif attr_name == "prime_meridian_name":
            return self._evaluate_prime_meridian_name(attr)

        elif attr_name == "projected_crs_name":
            return self._evaluate_projected_crs_name(attr)

        elif attr_name == "reference_ellipsoid_name":
            return self._evaluate_reference_ellipsoid_name(attr)

        elif attr_name == "towgs84":
            return self._evaluate_towgs84(attr)

        else:  # invoke method from 1.6, as these names are all still valid
            return super()._check_grid_mapping_attr_condition(
                attr,
                attr_name,
            )

    def _check_gmattr_existence_condition_geoid_name_geoptl_datum_name(self, var):
        """
        Check to see if both geoid_name and geopotential_datum_name exist as attributes
        for `var`. They should not.

        :param netCDF4.Variable var
        :rtype tuple
        :return two-tuple (bool, str)
        """

        msg = "Both geoid_name and geopotential_datum_name cannot exist"

        if ("geoid_name" in var.ncattrs()) and (
            "geopotential_datum_name" in var.ncattrs()
        ):
            return (False, msg)

        else:
            return (True, msg)

    def _check_gmattr_existence_condition_ell_pmerid_hdatum(self, var):
        """
        If one of reference_ellipsoid_name, prime_meridian_name, or
        horizontal_datum_name are defined as grid_mapping attributes,
        they must all be defined.

        :param netCDF4.Variable var
        :rtype tuple
        :return two-tuple (bool, str)
        """

        msg = (
            "If any of reference_ellipsoid_name, prime_meridian_name, "
            "or horizontal_datum_name are defined, all must be defined."
        )

        _ncattrs = set(var.ncattrs())

        if any(
            x in _ncattrs
            for x in [
                "reference_ellipsoid_name",
                "prime_meridian_name",
                "horizontal_datum_name",
            ]
        ) and (
            not {
                "reference_ellipsoid_name",
                "prime_meridian_name",
                "horizontal_datum_name",
            }.issubset(_ncattrs)
        ):
            return (False, msg)

        else:
            return (True, msg)

    def _exec_query_str_with_params(self, qstr, argtuple):
        """
        Execute a query string in a database connection with the given argument
        tuple. Return a result set.

        :param str qstr: desired query to be executed
        :param tuple argtuple: tuple of arguments to be supplied to query
        :rtype set
        """

        proj_db_path = os.path.join(pyproj.datadir.get_data_dir(), "proj.db")
        with sqlite3.connect(proj_db_path) as conn:
            return conn.execute(qstr, argtuple)

    def _evaluate_geographic_crs_name(self, val):
        """
        Evaluate the condition for the geographic_crs_name attribute.

        :param val: value to be tested
        :rtype tuple
        :return two-tuple of (bool, str)
        """

        query_str = (
            "SELECT 1 FROM geodetic_crs WHERE name = ? "
            "UNION ALL "  # need union in case contained in other tables
            "SELECT 1 FROM alias_name WHERE alt_name = ? "
            "AND table_name = 'geodetic_crs' LIMIT 1"
        )

        # try to find the value in the database
        res_set = self._exec_query_str_with_params(query_str, (val, val))

        # does it exist? if so, amt returned  be > 1
        return (
            len(res_set.fetchall()) > 0,
            "geographic_crs_name must correspond to a valid OGC WKT GEOGCS name",
        )

    def _evaluate_geoid_name(self, val):
        """
        Evaluate the condition for the geod_name attribute.

        :param val: value to be tested
        :rtype tuple
        :return two-tuple of (bool, str)
        """

        query_str = (
            "SELECT 1 FROM vertical_datum WHERE name = ? "
            "UNION ALL "
            "SELECT 1 FROM alias_name WHERE alt_name = ? "
            "AND table_name = 'vertical_datum' LIMIT 1"
        )

        # try to find the value in the database
        res_set = self._exec_query_str_with_params(query_str, (val, val))

        return (
            len(res_set.fetchall()) > 0,
            "geoid_name must correspond to a valid OGC WKT VERT_DATUM name",
        )

    def _evaluate_geopotential_datum_name(self, val):
        """
        Evaluate the condition for the geogpotential_datum_name attribute.

        :param val: value to be tested
        :rtype tuple
        :return two-tuple of (bool, str)
        """

        query_str = (
            "SELECT 1 FROM vertical_datum WHERE name = ? "
            "UNION ALL "
            "SELECT 1 FROM alias_name WHERE alt_name = ? "
            "AND table_name = 'vertical_datum' LIMIT 1"
        )

        # try to find the value in the database
        res_set = self._exec_query_str_with_params(query_str, (val, val))

        return (
            len(res_set.fetchall()) > 0,
            "geopotential_datum_name must correspond to a valid OGC WKT VERT_DATUM name",
        )

    def _evaluate_horizontal_datum_name(self, val):
        """
        Evaluate the condition for the horizontal_datum_name attribute.

        :param val: value to be tested
        :rtype tuple
        :return two-tuple of (bool, str)
        """

        return (
            val in horizontal_datum_names17,
            (
                "{} must be a valid Horizontal Datum Name; "
                "see https://github.com/cf-convention/cf-conventions/wiki/Mapping-from-CF-Grid-Mapping-Attributes-to-CRS-WKT-Elements."
            ),
        )

    def _evaluate_prime_meridian_name(self, val):
        """
        Evaluate the condition for the prime_meridian_name.

        :param val: value to be tested
        :rtype tuple
        :return two-tuple of (bool, str)
        """

        return (
            val in prime_meridian_names17,
            (
                "{} must be a valid Prime Meridian name; "
                "see https://github.com/cf-convention/cf-conventions/wiki/csv/prime_meridian.csv."
            ),
        )

    def _evaluate_projected_crs_name(self, val):
        """
        Evaluate the condition for the projected_crs attribute.

        :param val: value to be tested
        :rtype tuple
        :return two-tuple of (bool, str)
        """

        query_str = (
            "SELECT 1 FROM projected_crs WHERE name = ? "
            "UNION ALL "
            "SELECT 1 FROM alias_name WHERE alt_name = ? "
            "AND table_name = 'projected_crs' LIMIT 1"
        )

        # try to find the value in the database
        res_set = self._exec_query_str_with_params(query_str, (val, val))

        return (
            len(res_set.fetchall()) > 0,
            "projected_crs_name must correspond to a valid OGC WKT PROJCS name",
        )

    def _evaluate_reference_ellipsoid_name(self, val):
        """
        Evaluate the condition for the reference_ellipsoid_name attribute.

        :param val: value to be tested
        :rtype tuple
        :return two-tuple of (bool, str)
        """

        return (
            val in ellipsoid_names17,
            (
                "{} must be a valid Ellipsoid Name; "
                "see https://github.com/cf-convention/cf-conventions/wiki/csv/ellipsoid.csv."
            ),
        )

    def _evaluate_towgs84(self, val):
        """
        Evaluate the condition for the towgs84 attribute.

        :param val: value to be tested
        :rtype tuple
        :return two-tuple of (bool, str)
        """

        msg = (
            "towgs84 must be an array of length 3, 6, or 7 of double-precision"
            " and correspond to anm OGC WKT TOWGS84 node"
        )

        # if not numpy type, return false
        if not getattr(val, "dtype", None):
            return (False, msg)

        # must be double-precision array
        elif val.dtype != np.float64:
            return (False, msg)

        # must be of length 3, 6, or 7
        elif not val.shape:  # single value
            return (False, msg)

        elif val.size not in (3, 6, 7):
            return (False, msg)

        else:
            return (True, msg)

    def check_grid_mapping(self, ds):
        prev_return = super().check_grid_mapping(ds)
        grid_mapping_variables = cfutil.get_grid_mapping_variables(ds)
        for var_name in sorted(grid_mapping_variables):
            var = ds.variables[var_name]
            test_ctx = self.get_test_ctx(
                BaseCheck.HIGH,
                self.section_titles["5.6"],
                var.name,
            )

            # TODO: check cases where crs_wkt provides part of a necessary
            #       grid_mapping attribute, or where a grid_mapping attribute
            #       overrides what has been provided in crs_wkt.
            # attempt to parse crs_wkt if it is present
            if "crs_wkt" in var.ncattrs():
                crs_wkt = var.crs_wkt
                if not isinstance(crs_wkt, str):
                    test_ctx.messages.append("crs_wkt attribute must be a string")
                    test_ctx.out_of += 1
                else:
                    try:
                        pyproj.CRS.from_wkt(crs_wkt)
                    except pyproj.exceptions.CRSError as crs_error:
                        test_ctx.messages.append(
                            f"Cannot parse crs_wkt attribute to CRS using Proj4. Proj4 error: {str(crs_error)}",
                        )
                    else:
                        test_ctx.score += 1
                    test_ctx.out_of += 1

            # existence_conditions
            exist_cond_1 = (
                self._check_gmattr_existence_condition_geoid_name_geoptl_datum_name(var)
            )
            test_ctx.assert_true(exist_cond_1[0], exist_cond_1[1])
            exist_cond_2 = self._check_gmattr_existence_condition_ell_pmerid_hdatum(var)
            test_ctx.assert_true(exist_cond_2[0], exist_cond_2[1])

            # handle vertical datum related grid_mapping attributes
            vert_datum_attrs = {}
            possible_vert_datum_attrs = {"geoid_name", "geopotential_datum_name"}
            vert_datum_attrs = possible_vert_datum_attrs.intersection(var.ncattrs())
            len_vdatum_name_attrs = len(vert_datum_attrs)
            # check that geoid_name and geopotential_datum_name are not both
            # present in the grid_mapping variable
            if len_vdatum_name_attrs == 2:
                test_ctx.out_of += 1
                test_ctx.messages.append(
                    "Cannot have both 'geoid_name' and "
                    "'geopotential_datum_name' attributes in "
                    f"grid mapping variable '{var.name}'",
                )
            elif len_vdatum_name_attrs == 1:
                # should be one or zero attrs
                proj_db_path = os.path.join(pyproj.datadir.get_data_dir(), "proj.db")
                try:
                    with sqlite3.connect(proj_db_path) as conn:
                        v_datum_attr = next(iter(vert_datum_attrs))
                        v_datum_value = getattr(var, v_datum_attr)
                        v_datum_str_valid = self._process_v_datum_str(
                            v_datum_value,
                            conn,
                        )

                        invalid_msg = (
                            f"Vertical datum value '{v_datum_value}' for "
                            f"attribute '{v_datum_attr}' in grid mapping "
                            f"variable '{var.name}' is not valid"
                        )
                        test_ctx.assert_true(v_datum_str_valid, invalid_msg)
                except sqlite3.Error as e:
                    # if we hit an error, skip the check
                    warn(
                        "Error occurred while trying to query "
                        f"Proj4 SQLite database at {proj_db_path}: {str(e)}",
                        stacklevel=2,
                    )
            prev_return[var.name] = test_ctx.to_result()

        return prev_return

    def check_standard_name_deprecated_modifiers(self, ds):
        """
        Not a standard check in that it won't raise pass/fail values,
        but instead warns upon finding deprecated CF standard name modifiers.
        :param netCDF4.Dataset ds: netCDF dataset
        """
        deprecated_var_names = cfutil._find_standard_name_modifier_variables(ds, True)
        if deprecated_var_names:
            warn(
                f"Deprecated standard_name modifiers found on variables {deprecated_var_names}",
                stacklevel=2,
            )

    def _process_v_datum_str(self, v_datum_str, conn):
        vdatum_query = """SELECT 1 FROM alias_name WHERE
                            table_name = 'vertical_datum' AND
                            alt_name = ?
                                UNION ALL
                            SELECT 1 FROM vertical_datum WHERE
                            name = ?
                            LIMIT 1"""
        res_set = conn.execute(vdatum_query, (v_datum_str, v_datum_str))
        return len(res_set.fetchall()) > 0

    def _check_dimensionless_vertical_coordinate_1_7(
        self,
        ds,
        vname,
        deprecated_units,
        ret_val,
        dim_vert_coords_dict,
    ):
        """
        Check that a dimensionless vertical coordinate variable is valid under
        CF-1.7.

        :param netCDF4.Dataset ds: open netCDF4 dataset
        :param str name: variable name
        :param list ret_val: array to append Results to
        :rtype None
        """
        variable = ds.variables[vname]
        standard_name = getattr(variable, "standard_name", None)
        formula_terms = getattr(variable, "formula_terms", None)
        # Skip the variable if it's dimensional
        correct_computed_std_name_ctx = TestCtx(
            BaseCheck.MEDIUM,
            self.section_titles["4.3"],
        )
        # IMPLEMENTATION CONFORMANCE 4.3.3 REQUIRED
        correct_computed_std_name_ctx.assert_true(
            not (formula_terms is None and hasattr(variable, "computed_standard_name")),
            f"Variable {vname} should have formula_terms attribute when "
            "computed_standard_name attribute is defined",
        )
        if formula_terms is None and standard_name not in dim_vert_coords_dict:
            return

        # assert that the computed_standard_name is maps to the standard_name correctly
        _comp_std_name = dim_vert_coords_dict[standard_name][1]
        correct_computed_std_name_ctx.assert_true(
            getattr(variable, "computed_standard_name", None) in _comp_std_name,
            f"§4.3.3 The standard_name of `{vname}` must map to the correct computed_standard_name, `{sorted(_comp_std_name)}`",
        )
        ret_val.append(correct_computed_std_name_ctx.to_result())

    def check_dimensionless_vertical_coordinates(self, ds):
        """
        Check the validity of dimensionless coordinates under CF

        CF §4.3.2 The units attribute is not required for dimensionless
        coordinates.

        The standard_name attribute associates a coordinate with its definition
        from Appendix D, Dimensionless Vertical Coordinates. The definition
        provides a mapping between the dimensionless coordinate values and
        dimensional values that can positively and uniquely indicate the
        location of the data.

        A new attribute, formula_terms, is used to associate terms in the
        definitions with variables in a netCDF file.  To maintain backwards
        compatibility with COARDS the use of these attributes is not required,
        but is strongly recommended.

        :param netCDF4.Dataset ds: An open netCDF dataset
        :rtype: list
        :return: List of results
        """
        ret_val = []

        deprecated_units = ["level", "layer", "sigma_level"]

        # compose this function to use the results from the CF-1.6 check
        # and then extend it using a CF-1.7 addition
        ret_val.extend(
            self._check_dimensionless_vertical_coordinates(
                ds,
                deprecated_units,
                self._check_dimensionless_vertical_coordinate_1_6,
                dimless_vertical_coordinates_1_7,
            ),
        )

        ret_val.extend(
            self._check_dimensionless_vertical_coordinates(
                ds,
                deprecated_units,
                self._check_dimensionless_vertical_coordinate_1_7,
                dimless_vertical_coordinates_1_7,
            ),
        )

        return ret_val



from compliance_checker.cf.cf_1_9 import CF1_9Check


# no significant features for code implementation between CF 1.9 and CF 1.10
class CF1_10Check(CF1_9Check):
    _cc_spec_version = "1.10"
    _cc_url = "http://cfconventions.org/Data/cf-conventions/cf-conventions-1.10/cf-conventions.html"



#!/usr/bin/env python
import logging
import os
import sys
from collections import OrderedDict, defaultdict
from warnings import warn

import numpy as np
import regex

import compliance_checker.cf.util as cfutil
from compliance_checker.base import BaseCheck, BaseNCCheck, Result, TestCtx
from compliance_checker.cf import util
from compliance_checker.cf.appendix_d import no_missing_terms

logger = logging.getLogger(__name__)


class CFBaseCheck(BaseCheck):
    """
    CF Convention Checker Base
    """

    def __init__(self, options=None):
        # The compliance checker can be run on multiple datasets in a single
        # instantiation, so caching values has be done by the unique identifier
        # for each dataset loaded.

        # Each default dict is a key, value mapping from the dataset object to
        # a list of variables
        super().__init__(options)
        self._coord_vars = defaultdict(list)
        self._ancillary_vars = defaultdict(list)
        self._clim_vars = defaultdict(list)
        self._metadata_vars = defaultdict(list)
        self._boundary_vars = defaultdict(list)
        self._geophysical_vars = defaultdict(list)
        self._aux_coords = defaultdict(list)

        self._std_names = util.StandardNameTable()

        self.section_titles = {  # dict of section headers shared by grouped checks
            "1.2": "§1.2 Terminology",
            "2.1": "§2.1 Filename",
            "2.2": "§2.2 Data Types",
            "2.3": "§2.3 Naming Conventions",
            "2.4": "§2.4 Dimensions",
            "2.5": "§2.5 Variables",
            "2.5.1": "§2.5.1. Missing data, valid and actual range of data",
            "2.6": "§2.6 Attributes",
            "2.6.3": "§2.6.3 External Variables",
            "2.7.1": "§2.7.1. Scope",
            "3.1": "§3.1 Units",
            "3.2": "§3.2 Long Name",
            "3.3": "§3.3 Standard Name",
            "3.4": "§3.4 Ancillary Data",
            "3.5": "§3.5 Flags",
            "4": "§4 Coordinate Types",
            "4.1": "§4.1 Latitude Coordinate",
            "4.2": "§4.2 Longitude Coordinate",
            "4.3": "§4.3 Vertical Coordinate",
            "4.4": "§4.4 Time Coordinate",
            "4.4.1": "§4.4.1 Calendar",
            "4.5": "§4.5 Discrete Axis",
            "5": "§5 Coordinate Systems",
            "5.1": "§5.1 Independent Latitude, Longitude, Vertical, and Time Axes",
            "5.2": "§5.2 2-D Latitude, Longitude, Coordinate Variables",
            "5.3": "§5.3 Reduced Horizontal Grid",
            "5.4": "§5.4 Timeseries of Station Data",
            "5.5": "§5.5 Trajectories",
            "5.6": "§5.6 Horizontal Coordinate Reference Systems, Grid Mappings, Projections",
            "5.7": "§5.7 Scalar Coordinate Variables",
            "6.1": "§6.1 Labels",
            "6.2": "§6.2 Alternative Coordinates",
            "7.1": "§7.1 Cell Boundaries",
            "7.2": "§7.2 Cell Measures",
            "7.3": "§7.3 Cell Methods",
            "7.4": "§7.4 Climatological Statistics",
            "8.1": "§8.1 Packed Data",
            "8.2": "§8.2 Compression by Gathering",
            "9.1": "§9.1 Features and feature types",
            "9.2": "§9.2 Collections, instances, and elements",
            "9.3": "§9.3 Representations of Collections of features in data variables",
            "9.4": "§9.4 The featureType attribute",
            "9.5": "§9.5 Coordinates and metadata",
            "9.6": "§9.6 Missing Data",
        }

    ################################################################################
    # Helper Methods - var classifications, etc
    ################################################################################

    def setup(self, ds):
        """
        Initialize various special variable types within the class.
        Mutates a number of instance variables.

        :param netCDF4.Dataset ds: An open netCDF dataset
        """
        self.coord_vars = self._find_coord_vars(ds)
        self._find_aux_coord_vars(ds)
        self._find_ancillary_vars(ds)
        self._find_clim_vars(ds)
        self._find_boundary_vars(ds)
        self._find_metadata_vars(ds)
        self._find_cf_standard_name_table(ds)
        self._find_geophysical_vars(ds)
        coord_containing_vars = ds.get_variables_by_attributes(
            coordinates=lambda val: isinstance(val, str),
        )

        # coordinate data variables

        # Excerpt from "§1.3 Overview" on coordinate data
        # There are two methods used to identify variables that contain
        # coordinate data. The first is to use the NUG-defined "coordinate
        # variables." The use of coordinate variables is required for all
        # dimensions that correspond to one dimensional space or time
        # coordinates . In cases where coordinate variables are not applicable,
        # the variables containing coordinate data are identified by the
        # coordinates attribute.

        # first read in variables referred to in coordinates which exist
        # in the dataset
        self.coord_data_vars = set()
        for var in coord_containing_vars:
            for coord_var_name in var.coordinates.strip().split(" "):
                if coord_var_name in ds.variables:
                    self.coord_data_vars.add(coord_var_name)
        # then add in the NUG coordinate variables -- single dimension with
        # dimension name the same as coordinates
        self.coord_data_vars.update(self.coord_vars)

    def check_grid_mapping(self, ds):
        """
        5.6 When the coordinate variables for a horizontal grid are not
        longitude and latitude, it is required that the true latitude and
        longitude coordinates be supplied via the coordinates attribute. If in
        addition it is desired to describe the mapping between the given
        coordinate variables and the true latitude and longitude coordinates,
        the attribute grid_mapping may be used to supply this description.

        This attribute is attached to data variables so that variables with
        different mappings may be present in a single file. The attribute takes
        a string value which is the name of another variable in the file that
        provides the description of the mapping via a collection of attached
        attributes. This variable is called a grid mapping variable and is of
        arbitrary type since it contains no data. Its purpose is to act as a
        container for the attributes that define the mapping.

        The one attribute that all grid mapping variables must have is
        grid_mapping_name which takes a string value that contains the mapping's
        name. The other attributes that define a specific mapping depend on the
        value of grid_mapping_name. The valid values of grid_mapping_name along
        with the attributes that provide specific map parameter values are
        described in Appendix F, Grid Mappings.

        When the coordinate variables for a horizontal grid are longitude and
        latitude, a grid mapping variable with grid_mapping_name of
        latitude_longitude may be used to specify the ellipsoid and prime
        meridian.


        In order to make use of a grid mapping to directly calculate latitude
        and longitude values it is necessary to associate the coordinate
        variables with the independent variables of the mapping. This is done by
        assigning a standard_name to the coordinate variable. The appropriate
        values of the standard_name depend on the grid mapping and are given in
        Appendix F, Grid Mappings.

        :param netCDF4.Dataset ds: An open netCDF dataset
        :rtype: list
        :return: List of results
        """

        ret_val = OrderedDict()
        grid_mapping_variables = cfutil.get_grid_mapping_variables(ds)

        # Check the grid_mapping attribute to be a non-empty string and that its reference exists
        for variable in ds.get_variables_by_attributes(
            grid_mapping=lambda x: x is not None,
        ):
            grid_mapping = getattr(variable, "grid_mapping", None)
            defines_grid_mapping = self.get_test_ctx(
                BaseCheck.HIGH,
                self.section_titles["5.6"],
                variable.name,
            )
            defines_grid_mapping.assert_true(
                (isinstance(grid_mapping, str) and grid_mapping),
                f"{variable.name}'s grid_mapping attribute must be a "
                "space-separated non-empty string",
            )
            if isinstance(grid_mapping, str):
                # TODO (badams): refactor functionality to split functionality
                #                into requisite classes
                if ":" in grid_mapping and self._cc_spec_version >= "1.7":
                    colon_count = grid_mapping.count(":")
                    re_all = regex.findall(
                        r"(\w+):\s*((?:\w+\s+)*(?:\w+)(?![\w:]))",
                        grid_mapping,
                    )
                    if colon_count != len(re_all):
                        defines_grid_mapping.out_of += 1
                        defines_grid_mapping.messages.append(
                            "Could not consume entire grid_mapping expression, please check for well-formedness",
                        )
                    else:
                        for grid_var_name, coord_var_str in re_all:
                            defines_grid_mapping.assert_true(
                                grid_var_name in ds.variables,
                                f"grid mapping variable {grid_var_name} must exist in this dataset",
                            )
                            for ref_var in coord_var_str.split():
                                defines_grid_mapping.assert_true(
                                    ref_var in ds.variables,
                                    f"Coordinate-related variable {ref_var} referenced by grid_mapping variable {grid_var_name} must exist in this dataset",
                                )

                else:
                    for grid_var_name in grid_mapping.split():
                        defines_grid_mapping.assert_true(
                            grid_var_name in ds.variables,
                            f"grid mapping variable {grid_var_name} must exist in this dataset",
                        )
            ret_val[variable.name] = defines_grid_mapping.to_result()

        # Check the grid mapping variables themselves
        for grid_var_name in grid_mapping_variables:
            valid_grid_mapping = self.get_test_ctx(
                BaseCheck.HIGH,
                self.section_titles["5.6"],
                grid_var_name,
            )
            grid_var = ds.variables[grid_var_name]

            grid_mapping_name = getattr(grid_var, "grid_mapping_name", None)

            # Grid mapping name must be in appendix F
            valid_grid_mapping.assert_true(
                grid_mapping_name in self.grid_mapping_dict,
                f"{grid_mapping_name} is not a valid grid_mapping_name."
                + " See Appendix F for valid grid mappings",
            )

            # The self.grid_mapping_dict has a values of:
            # - required attributes
            # - optional attributes (can't check)
            # - required standard_names defined
            # - at least one of these attributes must be defined

            # We can't do any of the other grid mapping checks if it's not a valid grid mapping name
            if grid_mapping_name not in self.grid_mapping_dict:
                ret_val[grid_mapping_name] = valid_grid_mapping.to_result()
                continue

            grid_mapping = self.grid_mapping_dict[grid_mapping_name]
            required_attrs = grid_mapping[0]
            # Make sure all the required attributes are defined
            for req in required_attrs:
                valid_grid_mapping.assert_true(
                    hasattr(grid_var, req),
                    f"{req} is a required attribute for grid mapping {grid_mapping_name}",
                )

            # Make sure that exactly one of the exclusive attributes exist
            if len(grid_mapping) == 4:
                at_least_attr = grid_mapping[3]
                number_found = 0
                for attr in at_least_attr:
                    if hasattr(grid_var, attr):
                        number_found += 1
                valid_grid_mapping.assert_true(
                    number_found == 1,
                    f"grid mapping {grid_mapping_name}"
                    + "must define exactly one of these attributes: "
                    + "{}".format(" or ".join(at_least_attr)),
                )

            # Make sure that exactly one variable is defined for each of the required standard_names
            expected_std_names = grid_mapping[2]
            for expected_std_name in expected_std_names:
                found_vars = ds.get_variables_by_attributes(
                    standard_name=expected_std_name,
                )
                valid_grid_mapping.assert_true(
                    len(found_vars) == 1,
                    f"grid mapping {grid_mapping_name} requires exactly "
                    + "one variable with standard_name "
                    + f"{expected_std_name} to be defined",
                )

            ret_val[grid_var_name] = valid_grid_mapping.to_result()

        return ret_val

    def check_conventions_version(self, ds):
        """
        CF §2.6.1 the NUG defined global attribute Conventions to the string
        value "CF-<version_number>"; check the Conventions attribute contains
        the appropriate string.

        :param netCDF4.Dataset ds: An open netCDF dataset
        :rtype: compliance_checker.base.Result
        """

        valid = False
        reasoning = []
        correct_version_string = f"{self._cc_spec}-{self._cc_spec_version}".upper()
        if hasattr(ds, "Conventions"):
            conventions = regex.split(r",|\s+", getattr(ds, "Conventions", ""))
            for convention in conventions:
                if convention == correct_version_string:
                    valid = True
                    break
            else:
                reasoning = [
                    "§2.6.1 Conventions global attribute does not contain "
                    f'"{correct_version_string}"',
                ]
        else:
            valid = False
            reasoning = ["§2.6.1 Conventions field is not present"]
        return Result(
            BaseCheck.MEDIUM,
            valid,
            self.section_titles["2.6"],
            msgs=reasoning,
        )

    def _check_dimensionless_vertical_coordinates(
        self,
        ds,
        deprecated_units,
        version_specific_check,
        version_specific_dimless_vertical_coord_dict,
    ):
        """
        Check the validity of dimensionless coordinates under CF

        :param netCDF4.Dataset ds: An open netCDF dataset
        :param list deprecated_units: list of string names of deprecated units
        :param function version_specific_check: version-specific implementation to check dimensionless vertical coord
        :param dict version_specific_dimless_coord_dict: version-specific dict of dimensionless vertical coords and computed standard names
        :return: List of results
        """
        ret_val = []

        z_variables = cfutil.get_z_variables(ds)

        # call version-specific implementation
        for name in z_variables:
            version_specific_check(
                ds,
                name,
                deprecated_units,
                ret_val,
                version_specific_dimless_vertical_coord_dict,
            )

        return ret_val

    def _check_formula_terms(self, ds, coord, dimless_coords_dict):
        """
        Checks a dimensionless vertical coordinate contains valid formula_terms

        - formula_terms is a non-empty string
        - formula_terms matches regdimless_coords_dictx
        - every variable defined in formula_terms exists

        :param netCDF4.Dataset ds: An open netCDF dataset
        :rtype: compliance_checker.base.Result
        """
        variable = ds.variables[coord]
        standard_name = getattr(variable, "standard_name", None)
        formula_terms = getattr(variable, "formula_terms", None)
        valid_formula_terms = TestCtx(BaseCheck.HIGH, self.section_titles["4.3"])

        valid_formula_terms.assert_true(
            isinstance(formula_terms, str) and formula_terms,
            f"§4.3.2: {coord}'s formula_terms is a required attribute and must be a non-empty string"
            "",
        )
        # We can't check any more
        if not formula_terms:
            return valid_formula_terms.to_result()

        # check that the formula_terms are well formed and are present
        # The pattern for formula terms is always component: variable_name
        # the regex grouping always has component names in even positions and
        # the corresponding variable name in odd positions.
        poorly_formed_formula_terms = ("Attribute formula_terms is not well-formed",)
        matches = list(
            regex.finditer(
                r"(\w+):\s+(\w+)(?:\s+(?!$)|$)",
                variable.formula_terms,
            ),
        )
        if not matches:
            valid_formula_terms.add_failure(poorly_formed_formula_terms)
            return valid_formula_terms.to_result()

        terms = {m.group(1) for m in matches}
        # get the variables named in the formula terms and check if any
        # are not present in the dataset
        missing_vars = sorted({m.group(2) for m in matches} - set(ds.variables))
        missing_fmt = "The following variable(s) referenced in {}:formula_terms are not present in the dataset: {}"
        valid_formula_terms.assert_true(
            len(missing_vars) == 0,
            missing_fmt.format(coord, ", ".join(missing_vars)),
        )
        # try to reconstruct formula_terms by adding space in between the regex
        # matches.  If it doesn't exactly match the original, the formatting
        # of the attribute is incorrect
        reconstructed_formula = "".join(m.group(0) for m in matches)
        valid_formula_terms.assert_true(
            reconstructed_formula == formula_terms,
            "Attribute formula_terms is not well-formed",
        )

        valid_formula_terms.assert_true(
            standard_name in dimless_coords_dict,
            f"unknown standard_name '{standard_name}' for dimensionless vertical coordinate {coord}"
            "",
        )
        if standard_name not in dimless_coords_dict:
            return valid_formula_terms.to_result()

        valid_formula_terms.assert_true(
            no_missing_terms(standard_name, terms, dimless_coords_dict),
            f"{coord}'s formula_terms are invalid for {standard_name}, please see appendix D of CF 1.6"
            "",
        )

        return valid_formula_terms.to_result()

    def _check_grid_mapping_attr_condition(self, attr, attr_name, ret_val):
        """
        Evaluate a condition (or series of conditions) for a particular
        attribute. Designed to be overloaded in subclass implementations.

        :param attr: attribute to teset condition for
        :param str attr_name: name of the attribute
        :param list ret_val: list of results to append to
        :rtype None
        :return None
        """
        raise NotImplementedError

    def _dims_in_order(self, dimension_order):
        """
        :param list dimension_order: A list of axes
        :rtype: bool
        :return: Returns True if the dimensions are in order U*, T, Z, Y, X,
                 False otherwise
        """
        regx = regex.compile(r"^[^TZYX]*T?Z?Y?X?$")
        dimension_string = "".join(dimension_order)
        return regx.match(dimension_string) is not None

    def _parent_var_attr_type_check(self, attr_name, var, ctx):
        """
        Checks that an attribute has an equivalent value to a parent variable.
        Takes an attribute name, variable, and test context on which to operate.
        :param str attr_name: The name of the attribute to be checked
        :param netCDF4.Variable var: The variable against which to be checked
        :param compliance_checker.base.TestCtx ctx: The associated test context to modify
        :rtype None
        :return None
        """
        attr_val = var.getncattr(attr_name)

        if isinstance(attr_val, (str | bytes)):
            type_match = (var.dtype is str) or (var.dtype.kind == "S")
            val_type = type(attr_val)
        else:
            val_type = attr_val.dtype.type
            type_match = val_type == var.dtype.type

        ctx.assert_true(
            type_match,
            f"Attribute '{attr_name}' (type: {val_type}) and parent variable '{var.name}' (type: {var.dtype.type}) "
            "must have equivalent datatypes",
        )

    def _find_aux_coord_vars(self, ds, refresh=False):
        """
        Returns a list of auxiliary coordinate variables

        An auxiliary coordinate variable is any netCDF variable that contains
        coordinate data, but is not a coordinate variable (in the sense of the term
        defined by CF).

        :param netCDF4.Dataset ds: An open netCDF dataset
        :param bool refresh: if refresh is set to True, the cache is
                             invalidated.
        :rtype: list
        :return: List of variable names (str) that are defined to be auxiliary
                 coordinate variables.
        """
        ds_str = ds.__str__()
        if self._aux_coords.get(ds, None) and refresh is False:
            return self._aux_coords[ds_str]

        self._aux_coords[ds_str] = cfutil.get_auxiliary_coordinate_variables(ds)
        return self._aux_coords[ds_str]

    def _find_boundary_vars(self, ds, refresh=False):
        """
        Returns dictionary of boundary variables mapping the variable instance
        to the name of the variable acting as a boundary variable.

        :param netCDF4.Dataset ds: An open netCDF dataset
        :param bool refresh: if refresh is set to True, the cache is
                             invalidated.
        :rtype: list
        :return: A list containing strings with boundary variable names.
        """
        ds_str = ds.__str__()
        if self._boundary_vars.get(ds, None) and refresh is False:
            return self._boundary_vars[ds_str]

        self._boundary_vars[ds_str] = cfutil.get_cell_boundary_variables(ds)

        return self._boundary_vars[ds_str]

    def _find_ancillary_vars(self, ds, refresh=False):
        """
        Returns a list of variable names that are defined as ancillary
        variables in the dataset ds.

        An ancillary variable generally is a metadata container and referenced
        from other variables via a string reference in an attribute.

        - via ancillary_variables (3.4)
        - "grid mapping var" (5.6)
        - TODO: more?

        The result is cached by the passed in dataset object inside of this
        checker. Pass refresh=True to redo the cached value.

        :param netCDF4.Dataset ds: An open netCDF dataset
        :param bool refresh: if refresh is set to True, the cache is
                             invalidated.
        :rtype: list
        :return: List of variable names (str) that are defined as ancillary
                 variables in the dataset ds.
        """
        ds_str = ds.__str__()
        # Used the cached version if it exists and is not empty
        if self._ancillary_vars.get(ds, None) and refresh is False:
            return self._ancillary_vars[ds_str]

        # Invalidate the cache at all costs
        self._ancillary_vars[ds_str] = []

        for var in ds.variables.values():
            if hasattr(var, "ancillary_variables"):
                for anc_name in var.ancillary_variables.split(" "):
                    if anc_name in ds.variables:
                        self._ancillary_vars[ds_str].append(anc_name)

            if hasattr(var, "grid_mapping"):
                gm_name = var.grid_mapping
                if gm_name in ds.variables:
                    self._ancillary_vars[ds_str].append(gm_name)

        return self._ancillary_vars[ds_str]

    def _find_clim_vars(self, ds, refresh=False):
        """
        Returns a list of variables that are likely to be climatology variables based on CF §7.4

        :param netCDF4.Dataset ds: An open netCDF dataset
        :param bool refresh: if refresh is set to True, the cache is
                             invalidated.
        :rtype: list
        :return: A list containing strings with geophysical variable
                 names.
        """
        ds_str = ds.__str__()
        if self._clim_vars.get(ds, None) and refresh is False:
            return self._clim_vars[ds_str]

        climatology_variable = cfutil.get_climatology_variable(ds)
        if climatology_variable:
            self._clim_vars[ds_str].append(climatology_variable)

        return self._clim_vars[ds_str]

    def _find_cf_standard_name_table(self, ds):
        """
        Parse out the `standard_name_vocabulary` attribute and download that
        version of the cf standard name table.  If the standard name table has
        already been downloaded, use the cached version.  Modifies `_std_names`
        attribute to store standard names.  Returns True if the file exists and
        False if it fails to download.

        :param netCDF4.Dataset ds: An open netCDF dataset
        :rtype: bool
        """
        # Get the standard name vocab
        standard_name_vocabulary = getattr(ds, "standard_name_vocabulary", "")

        # Try to parse this attribute to get version
        version = None
        try:
            if "cf standard name table" in standard_name_vocabulary.lower():
                version = [
                    s.strip("(").strip(")").strip("v").strip(",")
                    for s in standard_name_vocabulary.split()
                ]
                # This assumes that table version number won't start with 0.
                version = [
                    s
                    for s in version
                    if s.isdigit() and len(s) <= 2 and not s.startswith("0")
                ]
                if len(version) > 1:
                    return False
                else:
                    try:
                        version = version[0]
                    except IndexError:
                        warn(
                            "Cannot extract CF standard name version number "
                            "from standard_name_vocabulary string",
                            stacklevel=2,
                        )
                        return False
            else:
                # Can't parse the attribute, use the packaged version
                return False
        # usually raised from .lower() with an incompatible (non-string)
        # data type
        except AttributeError:
            warn(
                "Cannot convert standard name table to lowercase.  This can "
                "occur if a non-string standard_name_vocabulary global "
                "attribute is supplied",
                stacklevel=2,
            )
            return False

        if version.startswith("v"):  # i.e 'v34' -> '34' drop the v
            version = version[1:]

        # If the packaged version is what we're after, then we're good
        if version == self._std_names._version:
            print(
                f"Using packaged standard name table v{version}",
                file=sys.stderr,
            )
            return False

        # Try to download the version specified
        try:
            data_directory = util.create_cached_data_dir()
            location = os.path.join(
                data_directory,
                f"cf-standard-name-table-test-{version}.xml",
            )
            # Did we already download this before?
            if not os.path.isfile(location):
                util.download_cf_standard_name_table(version, location)
                print(
                    f"Using downloaded standard name table v{version}",
                    file=sys.stderr,
                )
            else:
                print(
                    f"Using cached standard name table v{version} from {location}",
                    file=sys.stderr,
                )

            self._std_names = util.StandardNameTable(location)
            return True
        except Exception as e:
            # There was an error downloading the CF table. That's ok, we'll just use the packaged version
            warn(
                f"Problem fetching standard name table:\n{e}\n"
                f"Using packaged v{self._std_names._version}",
                stacklevel=2,
            )
            return False

    def _find_coord_vars(self, ds, refresh=False):
        """
        Returns a list of variable names that identify as coordinate variables.

        The result is cached by the passed in dataset object inside of this
        checker. Pass refresh=True to redo the cached value.

        :param netCDF4.Dataset ds: An open netCDF dataset
        :param bool refresh: if refresh is set to True, the cache is
                             invalidated.
        :rtype: list
        :return: A list of variables names (str) that are defined as coordinate
                 variables in the dataset ds.
        """
        ds_str = ds.__str__()
        if ds_str in self._coord_vars and refresh is False:
            return self._coord_vars[ds_str]

        self._coord_vars[ds_str] = cfutil.get_coordinate_variables(ds)

        return self._coord_vars[ds_str]

    def _find_geophysical_vars(self, ds, refresh=False):
        """
        Returns a list of geophysical variables.  Modifies
        `self._geophysical_vars`

        :param netCDF4.Dataset ds: An open netCDF dataset
        :param bool refresh: if refresh is set to True, the cache is
                             invalidated.
        :rtype: list
        :return: A list containing strings with geophysical variable
                 names.
        """
        ds_str = ds.__str__()
        if self._geophysical_vars.get(ds, None) and refresh is False:
            return self._geophysical_vars[ds_str]

        self._geophysical_vars[ds_str] = cfutil.get_geophysical_variables(ds)

        return self._geophysical_vars[ds_str]

    def _find_metadata_vars(self, ds, refresh=False):
        """
        Returns a list of netCDF variable instances for those that are likely metadata variables

        :param netCDF4.Dataset ds: An open netCDF dataset
        :param bool refresh: if refresh is set to True, the cache is
                             invalidated.
        :rtype: list
        :return:   List of variable names (str) that are likely metadata
                   variable candidates.

        """
        ds_str = ds.__str__()
        if self._metadata_vars.get(ds, None) and refresh is False:
            return self._metadata_vars[ds_str]

        self._metadata_vars[ds_str] = []
        for name, var in ds.variables.items():
            if name in self._find_ancillary_vars(ds) or name in self._find_coord_vars(
                ds,
            ):
                continue

            if name in (
                "platform_name",
                "station_name",
                "instrument_name",
                "station_id",
                "platform_id",
                "surface_altitude",
            ):
                self._metadata_vars[ds_str].append(name)

            elif getattr(var, "cf_role", "") != "":
                self._metadata_vars[ds_str].append(name)

            elif (
                getattr(var, "standard_name", None) is None and len(var.dimensions) == 0
            ):
                self._metadata_vars[ds_str].append(name)

        return self._metadata_vars[ds_str]

    def _get_coord_axis_map(self, ds):
        """
        Returns a dictionary mapping each coordinate to a letter identifier
        describing the _kind_ of coordinate.

        :param netCDF4.Dataset ds: An open netCDF dataset

        :rtype: dict
        :return: A dictionary with variable names mapped to axis abbreviations,
                 i.e. {'longitude': 'X', ... 'pressure': 'Z'}
        """
        expected = ["T", "Z", "Y", "X"]
        coord_vars = self._find_coord_vars(ds)
        coord_axis_map = {}

        # L - Unlimited Coordinates
        # T - Time coordinates
        # Z - Depth/Altitude Coordinate
        # Y - Y-Coordinate (latitude)
        # X - X-Coordinate (longitude)
        # A - Auxiliary Coordinate
        # I - Instance Coordinate

        time_variables = cfutil.get_time_variables(ds)
        lat_variables = cfutil.get_latitude_variables(ds)
        lon_variables = cfutil.get_longitude_variables(ds)
        z_variables = cfutil.get_z_variables(ds)

        for coord_name in coord_vars:
            coord_var = ds.variables[coord_name]
            axis = getattr(coord_var, "axis", None)
            standard_name = getattr(coord_var, "standard_name", None)

            # Unlimited dimensions must come first
            if ds.dimensions[coord_name].isunlimited():
                coord_axis_map[coord_name] = "L"
            # axis takes precedence over standard_name
            elif axis in expected:
                coord_axis_map[coord_name] = axis
            elif standard_name == "time":
                coord_axis_map[coord_name] = "T"
            elif standard_name == "longitude":
                coord_axis_map[coord_name] = "X"
            elif standard_name == "latitude":
                coord_axis_map[coord_name] = "Y"
            elif standard_name in ["height", "depth", "altitude"]:
                coord_axis_map[coord_name] = "Z"
            elif cfutil.is_compression_coordinate(ds, coord_name):
                coord_axis_map[coord_name] = "C"
            elif coord_name in time_variables:
                coord_axis_map[coord_name] = "T"
            elif coord_name in z_variables:
                coord_axis_map[coord_name] = "Z"
            elif coord_name in lat_variables:
                coord_axis_map[coord_name] = "Y"
            elif coord_name in lon_variables:
                coord_axis_map[coord_name] = "X"
            else:
                # mark the coordinate variable as unknown
                coord_axis_map[coord_name] = "U"

        for dimension in self._get_instance_dimensions(ds):
            if dimension not in coord_axis_map:
                coord_axis_map[dimension] = "I"

        # Dimensions of auxiliary coordinate variables will be marked with A.
        # This is useful to help determine if the dimensions are used like a
        # mapping from grid coordinates to physical lat/lon
        for coord_name in self._find_aux_coord_vars(ds):
            coord_var = ds.variables[coord_name]
            # Skip label auxiliary coordinates
            if hasattr(coord_var.dtype, "char") and coord_var.dtype.char == "S":
                continue
            elif coord_var.dtype == str:
                continue
            for dimension in coord_var.dimensions:
                if dimension not in coord_axis_map:
                    coord_axis_map[dimension] = "A"

        # If a dimension does not have a coordinate variable mark it as unknown
        # 'U'
        for dimension in ds.dimensions:
            if dimension not in coord_axis_map:
                coord_axis_map[dimension] = "U"

        return coord_axis_map

    def _get_coord_vars(self, ds):
        coord_vars = []
        for name, var in ds.variables.items():
            if (name,) == var.dimensions:
                coord_vars.append(name)
        return coord_vars

    def _get_dimension_order(self, ds, name, coord_axis_map):
        """
        Returns a list of strings corresponding to the named axis of the dimensions for a variable.

        Example::
            self._get_dimension_order(ds, 'temperature', coord_axis_map)
            --> ['T', 'Y', 'X']

        :param netCDF4.Dataset ds: An open netCDF dataset
        :param str name: Name of the variable
        :param dict coord_axis_map: A dictionary mapping each coordinate variable and dimension to a named axis

        :rtype: list
        :return: A list of strings corresponding to the named axis of the dimensions for a variable
        """

        retval = []
        variable = ds.variables[name]
        for dim in variable.dimensions:
            retval.append(coord_axis_map[dim])
        return retval

    def _get_instance_dimensions(self, ds):
        """
        Returns a list of dimensions marked as instance dimensions

        :param netCDF4.Dataset ds: An open netCDF dataset

        :rtype: list
        :returns: A list of variable dimensions
        """
        ret_val = []
        for variable in ds.get_variables_by_attributes(
            cf_role=lambda x: isinstance(x, str),
        ):
            if variable.ndim > 0:
                ret_val.append(variable.dimensions[0])
        return ret_val

    def _get_pretty_dimension_order(self, ds, name):
        """
        Returns a comma separated string of the dimensions for a specified
        variable

        :param netCDF4.Dataset ds: An open netCDF dataset
        :param str name: A string with a valid NetCDF variable name for the
                         dataset
        :rtype: str
        :return: A comma separated string of the variable's dimensions
        """
        dim_names = []
        for dim in ds.variables[name].dimensions:
            dim_name = dim
            if ds.dimensions[dim].isunlimited():
                dim_name += " (Unlimited)"
            dim_names.append(dim_name)
        return ", ".join(dim_names)

    def _get_pretty_dimension_order_with_type(self, ds, name, dim_types):
        """
        Returns a comma separated string of the dimensions for a specified
        variable of format "DIMENSIONS_NAME (DIMENSION_TYPE[, unlimited])"

        :param netCDF4.Dataset ds: An open netCDF dataset
        :param str name: A string with a valid NetCDF variable name for the
                         dataset
        :param list dim_types: A list of strings returned by
                               _get_dimension_order for the same "name"
        :rtype: str
        :return: A comma separated string of the variable's dimensions
        """
        dim_names = []
        for dim, dim_type in zip(
            ds.variables[name].dimensions,
            dim_types,
            strict=False,
        ):
            dim_name = f"{dim} ({dim_type}"
            if ds.dimensions[dim].isunlimited():
                dim_name += ", unlimited)"
            else:
                dim_name += ")"
            dim_names.append(dim_name)
        return ", ".join(dim_names)

    def _is_station_var(self, var):
        """
        Returns True if the NetCDF variable is associated with a station, False
        otherwise.

        :param netCDF4.Variable var: a variable in an existing NetCDF dataset
        :rtype: bool
        :return: Status of whether variable appears to be associated with a
                 station
        """

        if getattr(var, "standard_name", None) in (
            "platform_name",
            "station_name",
            "instrument_name",
        ):
            return True
        return False

    def _split_standard_name(self, standard_name):
        """
        Returns a tuple of the standard_name and standard_name modifier

        Nones are used to represent the absence of a modifier or standard_name

        :rtype: tuple
        :return: 2-tuple of standard_name and modifier as strings
        """

        if isinstance(standard_name, str) and " " in standard_name:
            return standard_name.split(" ", 1)
        # if this isn't a string, then it doesn't make sense to split
        # -- treat value as standard name with no modifier
        else:
            return standard_name, None

    def check_appendix_a(self, ds):
        """
        Validates a CF dataset against the contents of its Appendix A table for
        attribute types and locations. Returns a list of results with the
        outcomes of the Appendix A validation results against the existing
        attributes in the docstring.

        :param netCDF4.Variable var: a variable in an existing NetCDF dataset
        :param netCDF4.Dataset ds: An open netCDF dataset
        :rtype: list
        :return: A list of results corresponding to the results returned
        """
        # if 'enable_appendix_a_checks' isn't specified in the checks,
        # don't do anything on this check
        results = []
        if "enable_appendix_a_checks" not in self.options:
            return results
        possible_global_atts = set(ds.ncattrs()).intersection(self.appendix_a.keys())
        attr_location_ident = {
            "G": "global attributes",
            "C": "coordinate data",
            "D": "non-coordinate data",
        }

        def att_loc_print_helper(att_letter):
            """
            Returns a string corresponding to attr_location ident in
            human-readable form.  E.g. an input of 'G' will return
            "global attributes (G)"

            :param str att_letter: An attribute letter corresponding to the
                                   "Use" column in CF Appendix A
            :rtype: str
            :return: A string with a human-readable name followed by the input
                     letter specified
            """

            return "{} ({})".format(
                attr_location_ident.get(att_letter, "other"),
                att_letter,
            )

        def _att_loc_msg(att_loc):
            """
            Helper method for formatting an error message when an attribute
            appears in the improper location corresponding to the "Use" column
            in CF Appendix A.

            :param set att_loc: A set with the possible valid locations of the
                                attribute corresponding to the "Use" column
                                in CF Appendix A
            :rtype: str
            :return: A human-readable string with the possible valid locations
                     of the attribute
            """
            att_loc_len = len(att_loc)
            # this is a fallback in case an empty att_loc is passed
            # it generally should not occur
            valid_loc = "no locations in the dataset"
            loc_sort = sorted(att_loc)
            if att_loc_len == 1:
                valid_loc = att_loc_print_helper(loc_sort[0])
            elif att_loc_len == 2:
                valid_loc = f"{att_loc_print_helper(loc_sort[0])} and {att_loc_print_helper(loc_sort[1])}"
            # shouldn't be reached under normal circumstances, as any attribute
            # should be either G, C, or D but if another
            # category is added, this will be useful.
            else:
                valid_loc = (
                    ", ".join(loc_sort[:-1])
                    + f", and {att_loc_print_helper(loc_sort[-1])}"
                )
            return f"This attribute may only appear in {valid_loc}."

        for global_att_name in possible_global_atts:
            global_att = ds.getncattr(global_att_name)
            att_dict = self.appendix_a[global_att_name]
            att_loc = att_dict["attr_loc"]
            valid_loc_warn = _att_loc_msg(att_loc)
            if att_dict["cf_section"] is not None:
                subsection_test = ".".join(att_dict["cf_section"].split(".")[:2])

                section_loc = self.section_titles.get(
                    subsection_test,
                    att_dict["cf_section"],
                )
            else:
                section_loc = None
            test_ctx = TestCtx(BaseCheck.HIGH, section_loc)

            test_ctx.out_of += 1
            if "G" not in att_loc:
                test_ctx.messages.append(
                    f'[Appendix A] Attribute "{global_att_name}" should not be present in global (G) '
                    f"attributes. {valid_loc_warn}",
                )
            else:
                result = self._handle_dtype_check(global_att, global_att_name, att_dict)
                if not result[0]:
                    test_ctx.messages.append(result[1])
                else:
                    test_ctx.score += 1
            results.append(test_ctx.to_result())

        noncoord_vars = set(ds.variables) - set(self.coord_data_vars)
        for var_set, coord_letter in (
            (self.coord_data_vars, "C"),
            (noncoord_vars, "D"),
        ):
            for var_name in var_set:
                var = ds.variables[var_name]
                possible_attrs = set(var.ncattrs()).intersection(self.appendix_a.keys())
                for att_name in possible_attrs:
                    att_dict = self.appendix_a[att_name]
                    if att_dict["cf_section"] is not None:
                        subsection_test = ".".join(
                            att_dict["cf_section"].split(".")[:2],
                        )

                        section_loc = self.section_titles.get(
                            subsection_test,
                            att_dict["cf_section"],
                        )
                    else:
                        section_loc = None
                    test_ctx = TestCtx(BaseCheck.HIGH, section_loc, variable=var_name)
                    att_loc = att_dict["attr_loc"]
                    valid_loc_warn = _att_loc_msg(att_loc)
                    att = var.getncattr(att_name)
                    test_ctx.out_of += 1
                    if coord_letter not in att_loc:
                        test_ctx.messages.append(
                            f'[Appendix A] Attribute "{att_name}" should not be present in {att_loc_print_helper(coord_letter)} '
                            f'variable "{var_name}". {valid_loc_warn}',
                        )
                    else:
                        result = self._handle_dtype_check(att, att_name, att_dict, var)
                        if not result[0]:
                            test_ctx.messages.append(result[1])
                        else:
                            test_ctx.score += 1
                    results.append(test_ctx.to_result())

        return results

    def _check_attr_type(self, attr_name, attr_type, attribute, variable=None):
        """
        Check if an attribute `attr` is of the type `attr_type`. Upon getting
        a data type of 'D', the attr must have the same data type as the
        variable it is assigned to.

        Attributes designated type 'S' must be of type `str`. 'N' require
        numeric types, and 'D' requires the attribute type match the type
        of the variable it is assigned to.

        :param str attr_name: name of attr being checked (to format message)
        :param str attr_type: the correct type of the attribute
        :param attribute: attribute to check
        :param variable: if given, type should match attr
        :rtype tuple
        :return A two-tuple that contains pass/fail status as a boolean and
                a message string (or None if unset) as the second element.
        """

        if attr_type == "S":
            if not isinstance(attribute, str):
                return [False, f"{attr_name} must be a string"]
        else:
            # if it's not a string, it should have a numpy dtype
            underlying_dtype = getattr(attribute, "dtype", None)

            # TODO check for np.nan separately
            if underlying_dtype is None:
                return [False, f"{attr_name} must be a numeric type"]

            # both D and N should be some kind of numeric value
            is_numeric = np.issubdtype(underlying_dtype, np.number)
            if attr_type == "N":
                if not is_numeric:
                    return [False, f"{attr_name} must be a numeric type"]
            elif attr_type == "D":
                # TODO: handle edge case where variable is unset here
                temp_ctx = TestCtx()
                self._parent_var_attr_type_check(attr_name, variable, temp_ctx)
                var_dtype = getattr(variable, "dtype", None)
                if temp_ctx.messages:
                    return (
                        False,
                        f"{attr_name} must be numeric and must be equivalent to {var_dtype} dtype",
                    )
            else:
                # If we reached here, we fell off with an unrecognized type
                return (
                    False,
                    f"{attr_name} has unrecognized type '{attr_type}'",
                )
        # pass if all other possible failure conditions have been evaluated
        return (True, None)

    def _handle_dtype_check(self, attribute, attr_name, attr_dict, variable=None):
        """
        Helper function for Appendix A checks.

        :param attribute: The value of the attribute being checked
        :param str attr_name: The name of the attribute being processed
        :param dict attr_dict: The dict entry with type and attribute location
                               information corresponding to this attribute
        :param variable: if given, the variable whose type to check against
        :rtype: tuple
        :return: A two-tuple that contains pass/fail status as a boolean and
                 a message string (or None if unset) as the second element.
        """
        attr_type = attr_dict["Type"]
        if variable is None and "G" not in attr_dict["attr_loc"]:
            raise ValueError(
                "Non-global attributes must be associated with a " " variable",
            )
        attr_str = (
            f"Global attribute {attr_name}"
            if "G" in attr_dict["attr_loc"] and variable is None
            else f"Attribute {attr_name} in variable {variable.name}"
        )

        # check the type
        return_value = self._check_attr_type(attr_name, attr_type, attribute, variable)

        # if the second element is a string, format it
        if isinstance(return_value[1], str):
            return_value[1] = return_value[1].format(attr_str)

        # convert to tuple for immutability and return
        return tuple(return_value)


class CFNCCheck(BaseNCCheck, CFBaseCheck):
    """Inherits from both BaseNCCheck and CFBaseCheck to support
    checking netCDF datasets. Must inherit in this order, or certain
    attributes from BaseNCCheck (like supported_ds) will not be passed to
    CFNCCheck."""


appendix_a_base = {
    "Conventions": {"Type": "S", "attr_loc": {"G"}, "cf_section": None},
    "_FillValue": {"Type": "D", "attr_loc": {"D", "C"}, "cf_section": None},
    "add_offset": {"Type": "N", "attr_loc": {"D"}, "cf_section": "8.1"},
    "ancillary_variables": {"Type": "S", "attr_loc": {"D"}, "cf_section": "3.4"},
    "axis": {"Type": "S", "attr_loc": {"C"}, "cf_section": "4"},
    "bounds": {"Type": "S", "attr_loc": {"C"}, "cf_section": "7.1"},
    "calendar": {"Type": "S", "attr_loc": {"C"}, "cf_section": "4.4.1"},
    "cell_measures": {"Type": "S", "attr_loc": {"D"}, "cf_section": "7.2"},
    "cell_methods": {"Type": "S", "attr_loc": {"D"}, "cf_section": "7.3"},
    # cf_role type is "C" in document, which does not correspond
    # to types used, replaced with "S"
    "cf_role": {"Type": "S", "attr_loc": {"C"}, "cf_section": "9.5"},
    "climatology": {"Type": "S", "attr_loc": {"C"}, "cf_section": "7.4"},
    # comment was removed in this implementation
    "compress": {"Type": "S", "attr_loc": {"C"}, "cf_section": "8.2"},
    "coordinates": {"Type": "S", "attr_loc": {"D"}, "cf_section": "5"},
    # featureType type is "C" in document, which does not
    # correspond to types used, replaced with "S"
    "featureType": {"Type": "S", "attr_loc": {"G"}, "cf_section": "9.4"},
    "flag_masks": {"Type": "D", "attr_loc": {"D"}, "cf_section": "3.5"},
    "flag_meanings": {"Type": "S", "attr_loc": {"D"}, "cf_section": "3.5"},
    "flag_values": {"Type": "D", "attr_loc": {"D"}, "cf_section": "3.5"},
    "formula_terms": {"Type": "S", "attr_loc": {"C"}, "cf_section": "4.3.2"},
    "grid_mapping": {"Type": "S", "attr_loc": {"D"}, "cf_section": "5.6"},
    "history": {"Type": "S", "attr_loc": {"G"}, "cf_section": None},
    "institution": {"Type": "S", "attr_loc": {"G", "D"}, "cf_section": "2.6.2"},
    "leap_month": {"Type": "N", "attr_loc": {"C"}, "cf_section": "4.4.1"},
    "leap_year": {"Type": "N", "attr_loc": {"C"}, "cf_section": "4.4.1"},
    "long_name": {"Type": "S", "attr_loc": {"D", "C"}, "cf_section": "3.2"},
    "missing_value": {"Type": "D", "attr_loc": {"D", "C"}, "cf_section": "2.5.1"},
    "month_lengths": {"Type": "N", "attr_loc": {"C"}, "cf_section": "4.4.1"},
    "positive": {"Type": "S", "attr_loc": {"C"}, "cf_section": None},
    "references": {"Type": "S", "attr_loc": {"G", "D"}, "cf_section": "2.6.2"},
    "scale_factor": {"Type": "N", "attr_loc": {"D"}, "cf_section": "8.1"},
    "source": {"Type": "S", "attr_loc": {"G", "D"}, "cf_section": "2.6.2"},
    "standard_error_multiplier": {"Type": "N", "attr_loc": {"D"}, "cf_section": None},
    "standard_name": {"Type": "S", "attr_loc": {"D", "C"}, "cf_section": "3.3"},
    "title": {"Type": "S", "attr_loc": {"G"}, "cf_section": None},
    "units": {"Type": "S", "attr_loc": {"D", "C"}, "cf_section": "3.1"},
    "valid_max": {"Type": "N", "attr_loc": {"D", "C"}, "cf_section": None},
    "valid_min": {"Type": "N", "attr_loc": {"D", "C"}, "cf_section": None},
    "valid_range": {"Type": "N", "attr_loc": {"D", "C"}, "cf_section": None},
}



import json
import os
import sys
import traceback
from collections import OrderedDict
from contextlib import contextmanager

from compliance_checker.suite import CheckSuite


# Py 3.4+ has contextlib.redirect_stdout to redirect stdout to a different
# stream, but use this decorated function in order to redirect output in
# previous versions
@contextmanager
def stdout_redirector(stream):
    old_stdout = sys.stdout
    sys.stdout = stream
    try:
        yield
    finally:
        sys.stdout = old_stdout


class ComplianceChecker:
    """
    Compliance Checker runner class.

    Ties together the entire compliance checker framework, is used from
    the command line or can be used via import.
    """

    # Consider using __init__ instead of so many classmethods
    @classmethod
    def run_checker(
        cls,
        ds_loc,
        checker_names,
        verbose,
        criteria,
        skip_checks=None,
        include_checks=None,
        output_filename="-",
        output_format="text",
        options=None,
    ):
        """
        Static check runner.

        @param  ds_loc          Dataset location (url or file)
        @param  checker_names   List of string names to run, should match keys of checkers dict (empty list means run all)
        @param  verbose         Verbosity of the output (0, 1, 2)
        @param  criteria        Determines failure (lenient, normal, strict)
        @param  output_filename Path to the file for output
        @param  skip_checks     Names of checks to skip
        @param  include_checks  Names of checks to include
        @param  output_format   Format of the output(s)

        @returns                If the tests failed (based on the criteria)
        """
        all_groups = []
        cs = CheckSuite(options=options or {})
        # using OrderedDict is important here to preserve the order
        # of multiple datasets which may be passed in
        score_dict = OrderedDict()
        if not isinstance(ds_loc, str):
            locs = ds_loc
        # if single dataset, put in list
        else:
            locs = [ds_loc]

        # Make sure output format is a list
        if isinstance(output_format, str):
            output_format = [output_format]

        for loc in locs:  # loop through each dataset and run specified checks
            ds = cs.load_dataset(loc)

            score_groups = cs.run_all(ds, checker_names, include_checks, skip_checks)
            for group in score_groups.values():
                all_groups.append(group[0])
            # TODO: consider wrapping in a proper context manager instead
            if hasattr(ds, "close"):
                ds.close()

            if not score_groups:
                raise ValueError(
                    "No checks found, please check the name of the checker(s) and that they are installed",
                )
            else:
                score_dict[loc] = score_groups

        # define a score limit to truncate the output to the strictness level
        # specified by the user
        if criteria == "normal":
            limit = 2
        elif criteria == "strict":
            limit = 1
        elif criteria == "lenient":
            limit = 3

        for out_fmt in output_format:
            if out_fmt == "text":
                if output_filename == "-":
                    cls.stdout_output(cs, score_dict, verbose, limit)
                # need to redirect output from stdout since print functions are
                # presently used to generate the standard report output
                else:
                    if len(output_format) > 1:
                        # Update file name if needed
                        output_filename = f"{os.path.splitext(output_filename)[0]}.txt"
                    with open(output_filename, "w", encoding="utf-8") as f:
                        with stdout_redirector(f):
                            cls.stdout_output(cs, score_dict, verbose, limit)

            elif out_fmt == "html":
                # Update file name if needed
                if len(output_format) > 1 and output_filename != "-":
                    output_filename = f"{os.path.splitext(output_filename)[0]}.html"
                cls.html_output(cs, score_dict, output_filename, ds_loc, limit)

            elif out_fmt in {"json", "json_new"}:
                # Update file name if needed
                if len(output_format) > 1 and output_filename != "-":
                    output_filename = f"{os.path.splitext(output_filename)[0]}.json"
                cls.json_output(cs, score_dict, output_filename, ds_loc, limit, out_fmt)

            else:
                raise TypeError(f"Invalid format {out_fmt}")

            errors_occurred = cls.check_errors(score_groups, verbose)

        return (
            all(cs.passtree(groups, limit) for groups in all_groups),
            errors_occurred,
        )

    @classmethod
    def stdout_output(cls, cs, score_dict, verbose, limit):
        """
        Calls output routine to display results in terminal, including scoring.
        Goes to verbose function if called by user.

        @param cs           Compliance Checker Suite
        @param score_dict   Dict with dataset name as key, list of results as
                            value
        @param verbose      Integer value for verbosity level
        @param limit        The degree of strictness, 1 being the strictest, and going up from there.
        """

        for ds, score_groups in score_dict.items():
            for checker, (groups, _errors) in score_groups.items():
                score_list, points, out_of = cs.standard_output(
                    ds,
                    limit,
                    checker,
                    groups,
                )
                # send list of grouped result objects to stdout & reasoning_routine
                cs.standard_output_generation(
                    groups,
                    limit,
                    points,
                    out_of,
                    check=checker,
                )
        return groups

    @classmethod
    def html_output(cls, cs, score_dict, output_filename, ds_loc, limit):
        """
        Generates rendered HTML output for the compliance score(s)
        @param cs              Compliance Checker Suite
        @param score_groups    List of results
        @param output_filename The file path to output to
        @param ds_loc          List of source datasets
        @param limit           The degree of strictness, 1 being the strictest, and going up from there.
        """
        checkers_html = []
        for ds, score_groups in score_dict.items():
            for checker, (groups, _errors) in score_groups.items():
                checkers_html.append(cs.checker_html_output(checker, groups, ds, limit))

        html = cs.html_output(checkers_html)
        if output_filename == "-":
            print(html)
        else:
            with open(output_filename, "w", encoding="utf8") as f:
                f.write(html)

        return groups

    @classmethod
    def json_output(
        cls,
        cs,
        score_dict,
        output_filename,
        ds_loc,
        limit,
        output_type="json",
    ):
        """
        Generates JSON output for the ocmpliance score(s)
        @param cs              Compliance Checker Suite
        @param score_groups    List of results
        @param output_filename The file path to output to
        @param ds_loc          List of source datasets
        @param limit           The degree of strictness, 1 being the strictest,
                               and going up from there.
        @param output_type     Either 'json' or 'json_new'. json_new is the new
                               json output format that supports multiple datasets
        """
        results = {}
        # json output keys out at the top level by
        if len(score_dict) > 1 and output_type != "json_new":
            raise ValueError(
                "output_type must be set to 'json_new' if outputting multiple datasets to a single json file or stdout",
            )

        if output_type == "json":
            for ds, score_groups in score_dict.items():
                for checker, rpair in score_groups.items():
                    groups, errors = rpair
                    results[checker] = cs.dict_output(
                        checker,
                        groups,
                        ds,
                        limit,
                    )
        elif output_type == "json_new":
            for ds, score_groups in score_dict.items():
                results[ds] = {}
                for checker, rpair in score_groups.items():
                    groups, errors = rpair
                    results[ds][checker] = cs.dict_output(checker, groups, ds, limit)
        json_results = json.dumps(results, indent=2, ensure_ascii=False)

        if output_filename == "-":
            print(json_results)
        else:
            with open(output_filename, "w", encoding="utf8") as f:
                f.write(json_results)

        return groups

    @classmethod
    def check_errors(cls, score_groups, verbose):
        """
        Reports any errors (exceptions) that occurred during checking to stderr.
        Goes to verbose function if called by user.

        @param score_groups List of results
        @param verbose      Integer value for verbosity level
        """
        errors_occurred = False
        for checker, rpair in score_groups.items():
            errors = rpair[-1]
            if len(errors):
                errors_occurred = True
                print(
                    f"WARNING: The following exceptions occurred during the {checker} checker (possibly indicate compliance checker issues):",
                    file=sys.stderr,
                )
                for check_name, epair in errors.items():
                    print(
                        f"{checker}.{check_name}: {epair[0]}",
                        file=sys.stderr,
                    )

                    if verbose > 0:
                        traceback.print_tb(
                            epair[1].tb_next.tb_next,
                        )  # skip first two as they are noise from the running itself @TODO search for check_name
                        print(file=sys.stderr)

        return errors_occurred



import cftime
import numpy as np
import regex
from netCDF4 import Dataset

import compliance_checker.cf.util as cfutil
from compliance_checker.base import BaseCheck, Result, TestCtx
from compliance_checker.cf import util
from compliance_checker.cf.cf_1_8 import CF1_8Check
from compliance_checker.cf.util import (
    VariableReferenceError,
    get_coordinate_variables,
    reference_attr_variables,
)


class CF1_9Check(CF1_8Check):
    _cc_spec_version = "1.9"
    _cc_url = "http://cfconventions.org/Data/cf-conventions/cf-conventions-1.9/cf-conventions.html"
    _allowed_numeric_var_types = CF1_8Check._allowed_numeric_var_types.union(
        {np.ubyte, np.uint16, np.uint32, np.uint64, np.int64},
    )

    def __init__(self, options=None):
        super().__init__(options)
        self.section_titles.update({"5.8": "§5.8 Domain Variables"})

    def check_time_coordinate_variable_has_calendar(self, ds):
        """
        Ensure that time coordinate variables have a calendar attribute
        """
        ret_val = []
        for name in cfutil.get_time_variables(ds):
            # DRY: get rid of time coordinate variable boilerplate
            if name not in {var.name for var in util.find_coord_vars(ds)}:
                continue
            time_var = ds.variables[name]
            if not hasattr(time_var, "calendar") or not isinstance(
                time_var.calendar,
                str,
            ):
                result = Result(
                    BaseCheck.MEDIUM,
                    True,
                    self.section_titles["4.4.1"],
                    [
                        f'Time coordinate variable "{name}" should have a '
                        'string valued attribute "calendar"',
                    ],
                )
                ret_val.append(result)
                continue
            if time_var.calendar.lower() in {"gregorian", "julian", "standard"}:
                try:
                    reference_year = cftime.num2date(
                        0,
                        time_var.units,
                        time_var.calendar,
                        has_year_zero=True,
                    ).year
                # will fail on months, certain other time specifications
                except ValueError:
                    continue
                if reference_year == 0:
                    reasoning = (
                        f'For time variable "{time_var.name}", when using '
                        "the Gregorian or Julian calendars, the use of year "
                        "zero is not recommended. Furthermore, the use of year "
                        "zero to signify a climatological variable as in COARDS "
                        "is deprecated in CF."
                    )
                    result = Result(
                        BaseCheck.MEDIUM,
                        False,
                        self.section_titles["4.4.1"],
                        [reasoning],
                    )
                elif reference_year < 0:
                    reasoning = (
                        f'For time variable "{time_var.name}", when using '
                        "the Gregorian or Julian calendars, negative "
                        "years are not allowed."
                    )
                    result = Result(
                        BaseCheck.MEDIUM,
                        False,
                        self.section_titles["4.4.1"],
                        [reasoning],
                    )
                else:
                    return ret_val

                ret_val.append(result)
        return ret_val

    def check_time_coordinate(self, ds):
        prev_return = super().check_time_coordinate(ds)
        # adds check for < 60 seconds
        seconds_regex = regex.compile(
            r"\w+ since \d{1,4}-\d{1,2}-\d{1,2}[ T]"
            r"\d{1,2}:\d{1,2}:(?P<seconds>\d{1,2})",
        )
        for name in cfutil.get_time_variables(ds):
            # DRY: get rid of time coordinate variable boilerplate
            if name not in {var.name for var in util.find_coord_vars(ds)}:
                continue
            time_var = ds.variables[name]
            test_ctx = self.get_test_ctx(
                BaseCheck.HIGH,
                self.section_titles["4.4"],
                name,
            )
            try:
                match = regex.match(seconds_regex, time_var.units)
            except AttributeError:
                # not much can be done if there are no units
                continue
            # only check if seconds are present and in a parseable format
            if match:
                test_ctx.assert_true(
                    match.group("seconds") is None or int(match.group("seconds")) < 60,
                    f'Time coordinate variable "{name}" must have '
                    "units with seconds less than 60",
                )
                prev_return.append(test_ctx.to_result())
        return prev_return

    def check_domain_variables(self, ds: Dataset):

        results = []
        domain_valid = TestCtx(BaseCheck.MEDIUM, self.section_titles["5.8"])
        is_ragged_array_repr = cfutil.is_dataset_valid_ragged_array_repr_featureType(
            ds,
            getattr(ds, "featureType", ""),
        )
        if is_ragged_array_repr:
            domain_valid.out_of += 1
            ragged_array_dim_variable, ragged_attr_name = (
                cfutil.resolve_ragged_array_dimension(ds)
            )
            dim_name = getattr(ragged_array_dim_variable, ragged_attr_name)
            referenced_dim = reference_attr_variables(
                ds,
                dim_name,
                reference_type="dimensions",
            )
            if isinstance(referenced_dim, VariableReferenceError):
                domain_valid.messages.append(
                    f"Found ragged array variable {ragged_array_dim_variable.name}, "
                    f"but dimension {dim_name} referenced from {ragged_attr_name} does not exist in file",
                )

            coord_var_reference_failures = []
            for coord_var in reference_attr_variables(ds, dim_name, " ", "dimensions"):
                if isinstance(coord_var, VariableReferenceError):
                    coord_var_reference_failures.append(coord_var)
                    domain_valid.messages.append(
                        f"Referenced coordinate variable {coord_var} does not exist in file",
                    )
                    continue
                # TODO: check for label variables
                if not set(
                    util.get_possible_label_variable_dimensions(coord_var),
                ).issubset({referenced_dim}):
                    domain_valid.messages.append(
                        f"Found ragged array variable {ragged_array_dim_variable.name}, "
                        f"but dimension {dim_name} referenced from {ragged_attr_name} does not exist in file",
                    )
                else:
                    domain_valid.score += 1
        else:
            # IMPLICIT CONFORMANCE REQUIRED 1/4
            # Domain variables must have a dimensions attribute
            for domain_var in (
                ds.variables[var]
                for var in ds.variables
                if "dimensions" in ds.variables[var].ncattrs()
            ):
                # store NetCDF variable attribute named "dimensions"
                dim_nc_attr = domain_var.getncattr("dimensions")
                # Every variable should have dimensionality (not to be confused with
                # dimensions NetCDF attribute here) even if scalar/empty, but non-empty
                # dimensions aren't allowed for domain variables

                domain_valid.out_of += 3
                if domain_var.dimensions:
                    domain_valid.messages.append(
                        "Domain variable "
                        f"{domain_var.name} should not have non-scalar/"
                        "non-empty variable dimensions",
                    )
                else:
                    domain_valid.score += 1

                # IMPLICIT CONFORMANCE REQUIRED 2/4
                # Aforementioned dimensions attribute is comprised of space separated
                # dimension names which must exist in the file

                domain_dims, dim_errors = reference_attr_variables(
                    ds,
                    dim_nc_attr,
                    " ",
                    "dimensions",
                )
                if dim_errors:
                    errors_str = ", ".join(dim_errors)
                    domain_valid.messages.append(
                        "Could not find the following "
                        "dimensions referenced in "
                        "dimensions attribute from "
                        "domain variable "
                        f"{domain_var.name}: {errors_str}",
                    )
                else:
                    domain_valid.score += 1
                if "coordinates" in domain_var.ncattrs():
                    domain_coord_vars, domain_coord_var_errors = (
                        reference_attr_variables(ds, domain_var.coordinates, " ")
                    )
                    if domain_coord_var_errors:
                        errors_str = ", ".join(
                            err.name for err in domain_coord_var_errors
                        )
                        domain_valid.messages.append(
                            "Could not find the following "
                            "variables referenced in "
                            "coordinates attribute from "
                            "domain variable "
                            f"{domain_var.name}: {errors_str}",
                        )
                    else:
                        domain_valid.score += 1
                        for coord_var in domain_coord_vars:
                            if coord_var not in get_coordinate_variables(ds):
                                continue
                            domain_valid.out_of += 1
                            domain_dims_names = {var.name for var in domain_dims}
                            variable_dim = util.get_possible_label_variable_dimensions(
                                coord_var,
                            )
                            if not (
                                set(
                                    util.get_possible_label_variable_dimensions(
                                        coord_var,
                                    ),
                                ).issubset(domain_dims_names)
                            ):
                                domain_valid.messages.append(
                                    "Could not find the following "
                                    "variables referenced in "
                                    "coordinates attribute from "
                                    "domain variable "
                                    f"{variable_dim}: {domain_dims_names}",
                                )
                            else:
                                domain_valid.score += 1

                # not in conformance docs, but mentioned as recommended anyways
                domain_valid.out_of += 1
                long_name = getattr(domain_var, "long_name", None)
                if long_name is None or not isinstance(long_name, str):
                    domain_valid.messages.append(
                        f"For domain variable {domain_var.name} "
                        f"it is recommended that attribute long_name be present and a string",
                    )
                else:
                    domain_valid.score += 1

                appendix_a_not_recommended_attrs = []
                for attr_name in domain_var.ncattrs():
                    if (
                        attr_name in self.appendix_a
                        and "D" not in self.appendix_a[attr_name]["attr_loc"]
                    ):
                        appendix_a_not_recommended_attrs.append(attr_name)

                domain_valid.out_of += 1
                if appendix_a_not_recommended_attrs:
                    domain_valid.messages.append(
                        f"The following attributes appear in variable {domain_var.name} "
                        "and CF Appendix A, but are not for use in domain variables: "
                        f"{appendix_a_not_recommended_attrs}",
                    )
                else:
                    # no errors occurred
                    domain_valid.score += 1

                # IMPLEMENTATION CONFORMANCE 5.8 REQUIRED 4/4
                # variables named by domain variable's cell_measures attributes must themselves be a subset
                # of dimensions named by domain variable's dimensions NetCDF attribute
                if hasattr(domain_var, "cell_measures"):
                    cell_measures_var_names = regex.findall(
                        r"\b(?:area|volume):\s+(\w+)",
                        domain_var.cell_measures,
                    )
                    # check exist
                    for var_name in cell_measures_var_names:
                        try:
                            cell_measures_variable = ds.variables[var_name]
                        except ValueError:
                            # TODO: what to do here?
                            continue
                        domain_valid.assert_true(
                            set(cell_measures_variable.dimensions).issubset(
                                {dd.name for dd in domain_dims},
                            ),
                            "Variables named in the cell_measures attributes must have dimensions with "
                            "values that are a subset of the referring domain variable's dimensions attribute",
                        )

            results.append(domain_valid.to_result())

        return results
